"""
Evaluators for binary classification tests (R=2).

Classes:
    SupervisedEvaluation
    ErrorIndependentEvaluation
    MajorityVotingEvaluation

Functions:

Misc variables:

"""

import math, itertools
import sympy
from fractions import Fraction
from typing_extensions import Iterable

from ntqr.r2.datasketches import Label, Votes
from ntqr.r2.datasketches import (
    classifier_label_votes,
    classifiers_labels_votes,
)
from ntqr.r2.datasketches import TrioLabelVoteCounts, TrioVoteCounts
from ntqr.r2.examples import uciadult_label_counts


class SupervisedEvaluation:
    """Evaluation for experiments where the true labels are known."""

    # While the classes are transitioned to ensembles of arbitrary size,
    # this class is hard-wired for three classifiers
    vote_patterns = list(itertools.product(*["ab" for i in range(3)]))
    pairs = ((0, 1), (0, 2), (1, 2))

    def __init__(self, label_counts: TrioLabelVoteCounts):
        self.label_counts = label_counts

        self.evaluation_exact = {
            "prevalence": self.prevalences(),
            "accuracy": [
                {
                    label: self.classifier_label_accuracy(classifier, label)
                    for label in ("a", "b")
                }
                for classifier in range(3)
            ],
            "pair_correlation": {
                pair: {
                    label: self.pair_label_error_correlation(pair, label)
                    for label in ("a", "b")
                }
                for pair in self.pairs
            },
            "3_way_correlation": {
                trio: {
                    label: self.three_way_label_error_correlation(trio, label)
                    for label in ("a", "b")
                }
                for trio in ((0, 1, 2),)
            },
        }

        self.evaluation_float = {
            "prevalence": {
                label: float(val)
                for label, val in self.evaluation_exact["prevalence"].items()
            },
            "accuracy": [
                {label: float(val) for label, val in cdict.items()}
                for cdict in self.evaluation_exact["accuracy"]
            ],
            "pair_correlation": {
                pair: {label: float(val) for label, val in corrs.items()}
                for pair, corrs in self.evaluation_exact[
                    "pair_correlation"
                ].items()
            },
        }

    def prevalences(self):
        """
        Calculate the prevalences of the two labels.

        Returns
        -------
        Mapping[Label, Fraction]
            Mapping from labels to percentage of appearance in the test.

        """
        test_sizes = self.label_counts.test_sizes
        total = sum(test_sizes.values())
        return {
            "a": sympy.Rational(test_sizes["a"], total),
            "b": sympy.Rational(test_sizes["b"], total),
        }

    def classifier_label_accuracy(self, classifier: int, label: Label):
        """Compute classifier label accuracy."""
        test_size = self.label_counts.test_sizes[label]
        classifier_votes = classifier_label_votes(
            classifier, label, self.vote_patterns
        )
        label_counts = self.label_counts[label]
        correct_counts = [label_counts[votes] for votes in classifier_votes]
        return sympy.Rational(sum(correct_counts), test_size)

    def other_label(self, label: Label):
        """Return the other binary classification label given label."""
        if label == "a":
            o_label = "b"
        else:
            o_label = "a"
        return o_label

    def pair_label_error_correlation(self, pair, label):
        """Calculate the label error correlation a classifier pair."""
        test_size = self.label_counts.test_sizes[label]
        classifier_accuracies = [
            self.classifier_label_accuracy(classifier, label)
            for classifier in pair
        ]
        label_counts = self.label_counts[label]
        o_label = self.other_label(label)
        return (
            # They are both correct
            (1 - classifier_accuracies[0])
            * (1 - classifier_accuracies[1])
            * sum(
                [
                    label_counts[votes]
                    for votes in classifiers_labels_votes(
                        pair, (label, label), self.vote_patterns
                    )
                ]
            )
            +
            # C_i is correct, C_j is incorrect
            (1 - classifier_accuracies[0])
            * (0 - classifier_accuracies[1])
            * sum(
                [
                    label_counts[votes]
                    for votes in classifiers_labels_votes(
                        pair, (label, o_label), self.vote_patterns
                    )
                ]
            )
            +
            # C_i is incorrect, C_j is correct
            (0 - classifier_accuracies[0])
            * (1 - classifier_accuracies[1])
            * sum(
                [
                    label_counts[votes]
                    for votes in classifiers_labels_votes(
                        pair, (o_label, label), self.vote_patterns
                    )
                ]
            )
            +
            # C_i and C_j are incorrect
            (0 - classifier_accuracies[0])
            * (0 - classifier_accuracies[1])
            * sum(
                [
                    label_counts[votes]
                    for votes in classifiers_labels_votes(
                        pair, (o_label, o_label), self.vote_patterns
                    )
                ]
            )
        ) / test_size

    def three_way_label_error_correlation(self, triplet, label):
        """Calculate the label error correlation a classifier pair."""
        test_size = self.label_counts.test_sizes[label]
        classifier_accuracies = [
            self.classifier_label_accuracy(classifier, label)
            for classifier in triplet
        ]
        label_counts = self.label_counts[label]
        o_label = self.other_label(label)
        return (
            # They are all correct
            (1 - classifier_accuracies[0])
            * (1 - classifier_accuracies[1])
            * (1 - classifier_accuracies[2])
            * sum(
                [
                    label_counts[votes]
                    for votes in classifiers_labels_votes(
                        triplet, (label, label, label), self.vote_patterns
                    )
                ]
            )
            +
            # C_i is correct, C_j is correct, C_k is incorrect
            (1 - classifier_accuracies[0])
            * (1 - classifier_accuracies[1])
            * (0 - classifier_accuracies[2])
            * sum(
                [
                    label_counts[votes]
                    for votes in classifiers_labels_votes(
                        triplet, (label, label, o_label), self.vote_patterns
                    )
                ]
            )
            +
            # C_i is correct, C_j is incorrect, C_k is correct
            (1 - classifier_accuracies[0])
            * (0 - classifier_accuracies[1])
            * (1 - classifier_accuracies[2])
            * sum(
                [
                    label_counts[votes]
                    for votes in classifiers_labels_votes(
                        triplet, (label, o_label, label), self.vote_patterns
                    )
                ]
            )
            +
            # C_i is incorrect, C_j is correct, C_k is correct
            (0 - classifier_accuracies[0])
            * (1 - classifier_accuracies[1])
            * (1 - classifier_accuracies[2])
            * sum(
                [
                    label_counts[votes]
                    for votes in classifiers_labels_votes(
                        triplet, (o_label, label, label), self.vote_patterns
                    )
                ]
            )
            +
            # C_i is correct, C_j is incorrect, C_k is incorrect
            (1 - classifier_accuracies[0])
            * (0 - classifier_accuracies[1])
            * (0 - classifier_accuracies[2])
            * sum(
                [
                    label_counts[votes]
                    for votes in classifiers_labels_votes(
                        triplet, (label, o_label, o_label), self.vote_patterns
                    )
                ]
            )
            +
            # C_i is incorrect, C_j is correct, C_k is incorrect
            (0 - classifier_accuracies[0])
            * (1 - classifier_accuracies[1])
            * (0 - classifier_accuracies[2])
            * sum(
                [
                    label_counts[votes]
                    for votes in classifiers_labels_votes(
                        triplet, (o_label, label, o_label), self.vote_patterns
                    )
                ]
            )
            +
            # C_i is incorrect, C_j is incorrect, C_k is correct
            (0 - classifier_accuracies[0])
            * (0 - classifier_accuracies[1])
            * (1 - classifier_accuracies[2])
            * sum(
                [
                    label_counts[votes]
                    for votes in classifiers_labels_votes(
                        triplet, (o_label, o_label, label), self.vote_patterns
                    )
                ]
            )
            +
            # C_i is incorrect, C_j is incorrect, C_k is incorrect
            (0 - classifier_accuracies[0])
            * (0 - classifier_accuracies[1])
            * (0 - classifier_accuracies[2])
            * sum(
                [
                    label_counts[votes]
                    for votes in classifiers_labels_votes(
                        triplet,
                        (o_label, o_label, o_label),
                        self.vote_patterns,
                    )
                ]
            )
        ) / test_size


class ErrorIndependentEvaluation:
    """
    Evaluate three binary classifiers assuming they are error independent.

    Returns
    -------
    Absent labeled data, there are two logically consistent solutions
    given only their decision voting frequencies. For binary classification,
    this means that there are 2 possible points in evaluation space that
    can possibly explain the test results. The ground truth evaluation is
    one of these points -- if the assumption of error independence is true.

    The exact algebraic results have a unique virtue that few alarm systems
    have - it can warn about the failures of its own assumption of error
    independence. If the two possible solutions for the 'a' label prevalence
    return an unresolved integer square root - the classifiers are error
    correlated in the evaluation.

    In version 0.2 the math needed to take handle the almost certain detection
    of error correlation will be added. It is already being built as
    can be seen in ntqr.r2.postulates file where the postulates related to
    computing the error correlation have been expressed using the SymPy
    package.

    Warnings:
    ---------
    A. The ntqr package uses a notion of 'error independence' that is
    different than the one most familiar in the ML/AI community. There are
    many notions of independence in mathematics. In the context of ML/AI
    papers/discussions, the term 'error independence' is taken to be:
    A.1. Functional independence of distributions: P(x, y) = P(x)P(y)
    The one used in the ntqr package is sample defined since there is no
    probability theory used in its logic. For that reason, you must define a
    set of error correlation parameters. 'Error independence' in the ntqr
    package means:
    A.2. pair_label_correlations = 0, trio_label_correlations = 0, ...
    It is best to think of 'error independence' in the ntqr package as a
    property that belongs to the classifiers AND the test they took.

    B. This class currently assumes that the observed classifier
    vote counts supplied by the user are not fake. The set of all valid
    observations from a classification test is much smaller than the set
    of all sets of eight positive integers. Future versions of the ntqr
    package will implement the algebraic geometry needed to detect when
    TrioVoteCounts objects are not explainable as observations from a
    classification test.

    The error independent solution can fail if, in fact, the classifiers
    are highly correlated on the test being evaluated. Tests can fail.
    Future versions will have implemented the exceptions.
        1. PrevalenceImaginaryException
        2. NoSolutionException

    The PrevalenceImaginaryException is a iron-clad detection of highly
    correlated classifiers. Its main utility will be in "warning light"
    applications in AI safety.

    The NoSolutionException means that NO independent system can possibly
    explain the observations. There are two different reasons for this -
    higher error correlations, or the data sketch is fake. Distinguishing
    between the two comes down to the same computation of error correlation.
    """

    def __init__(self, vote_counts: TrioVoteCounts):
        self.vote_counts = vote_counts
        self.vote_frequencies = self.vote_counts.to_frequencies_exact()

        # Two solutions are possible
        self.evaluation_exact = [
            {
                "prevalence": {"a": a_prevalence, "b": 1 - a_prevalence},
                "accuracy": [
                    {
                        "a": self.classifier_a_label_accuracy(
                            classifier, a_prevalence
                        ),
                        "b": self.classifier_b_label_accuracy(
                            classifier, a_prevalence
                        ),
                    }
                    for classifier in range(3)
                ],
            }
            for a_prevalence in self.alpha_prevalence_estimates()
        ]

        self.evaluation_float = [
            {
                "prevalence": {
                    label: float(val)
                    for label, val in sol_dict["prevalence"].items()
                },
                "accuracy": [
                    {
                        label: float(val)
                        for label, val in classifier_dict.items()
                    }
                    for classifier_dict in sol_dict["accuracy"]
                ],
            }
            for sol_dict in self.evaluation_exact
        ]

    def alpha_prevalence_quadratic_terms(self):
        """
        Calculate the coefficients of the 'a' label prevalence quadratic.

        If the quadratic is represented as:
            a * (P_a)**2 + b * P_a + c
        then,
            a = terms[2], b = terms[1], c = terms[0].
        The quadratic is written in the 'standard' way seen in algebra
        textbooks. Be careful to not mistake the 'a' or 'b' coefficients
        described here with the two labels being used for classification -
        currently implemented as ('a', 'b').
        """

        pfmds = self.vote_counts.label_pairs_frequency_moments("b")
        vote_frequencies = self.vote_frequencies
        fbbb = vote_frequencies[("b", "b", "b")]
        diff1 = fbbb - self.vote_counts.trio_frequency_moment()
        prodFDs = math.prod(pfmds.values())
        term_coefficients = {
            2: diff1**2 + 4 * prodFDs,
            1: -(diff1**2 + 4 * prodFDs),
            0: prodFDs,
        }

        return term_coefficients

    def alpha_prevalence_estimates(self):
        """Calculate the prevalence of the alpha label.

        Since the quadratic equation has ordered solutions by
        the plus/minus operations, we arbitrarily return the 'a' label
        less than 50% solution first."""
        coeffs = self.alpha_prevalence_quadratic_terms()
        a = coeffs[2]
        c = coeffs[0]
        sqrTerm = sympy.sqrt(1 - 4 * c / a) / 2
        prev_expr = [
            sympy.Rational(1, 2) - sqrTerm,
            sympy.Rational(1, 2) + sqrTerm,
        ]
        return [sympy.simplify(expr) for expr in prev_expr]

    def classifier_a_label_accuracy(self, classifier: int, a_prevalence):
        """
        Calculate classifier 'a' label accuracies.

        Parameters
        ----------
        classifier : int
            One of (0, 1, 2).
        a_prevalence: Sympy expression

        Returns
        -------
        The a label accuracy given the a_prevalence value
        """
        # Assuming the linear equation form a + b * P_a + c * P_{i, a}

        classifier_freqs = [
            self.vote_counts.classifier_label_frequency(classifier, "b")
            for classifier in range(3)
        ]
        fbbb = self.vote_frequencies[("b", "b", "b")]
        fijk = self.vote_counts.trio_frequency_moment()
        freqDiff = fbbb - fijk

        pair_moments = self.vote_counts.label_pairs_frequency_moments("b")
        prodFDs = math.prod(pair_moments.values())
        match classifier:
            case 0:
                other_moment = pair_moments[(1, 2)]
            case 1:
                other_moment = pair_moments[(0, 2)]
            case 2:
                other_moment = pair_moments[(0, 1)]
            case _:
                pass
        a_coeff = (
            freqDiff * freqDiff
            - other_moment * (1 - classifier_freqs[classifier]) * freqDiff
            + 2 * prodFDs
        )

        b_coeff = -4 * prodFDs - freqDiff**2
        c_coeff = other_moment * freqDiff

        return sympy.simplify((-a_coeff - b_coeff * a_prevalence) / c_coeff)

    def classifier_b_label_accuracy(self, classifier: int, a_prevalence):
        """
        Calculate classifier 'b' label accuracies.

        Parameters
        ----------
        classifier : int
            One of (0, 1, 2).

        Returns
        -------
        Two possible logically consistent estimates for P_{i,b} given the
        test error independence assumption.
        """
        # Assuming the linear equation form a + b * P_a + c * P_{i, b}

        classifier_freqs = [
            self.vote_counts.classifier_label_frequency(classifier, "b")
            for classifier in range(3)
        ]
        fbbb = self.vote_frequencies[("b", "b", "b")]
        fijk = self.vote_counts.trio_frequency_moment()
        freqDiff = fbbb - fijk

        pair_moments = self.vote_counts.label_pairs_frequency_moments("b")
        prodFDs = math.prod(pair_moments.values())
        match classifier:
            case 0:
                other_moment = pair_moments[(1, 2)]
            case 1:
                other_moment = pair_moments[(0, 2)]
            case 2:
                other_moment = pair_moments[(0, 1)]
            case _:
                pass
        a_coeff = (
            classifier_freqs[classifier] * other_moment * freqDiff
            - 2 * prodFDs
        )

        b_coeff = 4 * prodFDs + freqDiff**2
        c_coeff = -other_moment * freqDiff

        return sympy.simplify((-a_coeff - b_coeff * a_prevalence) / c_coeff)


class MajorityVotingEvaluation:
    """
    Evaluate three binary classifiers using majority voting.

    Majority voting can be used to carry out evaluation algebraically.
    Typically, majority voting is used with the assumption that the
    crowd is always right. In the context of safety, however, that the
    crowd is always wrong is an equally valid a-priori assumption. Hence,
    this class returns TWO evaluations. The first assuming the crowd is
    always right and the second assuming they are always wrong.
    Its main virtue is that it is simple and rock solid - always returns
    logically consistent evaluations.
    """

    vote_patterns = list(itertools.product(*["ab" for i in range(3)]))

    def __init__(self, vote_counts: TrioVoteCounts):
        """
        Initialize data structures.

        TODO: Implement detection that the TrioVoteCounts are not generated
        by a classification test.

        Parameters
        ----------
        vote_counts : TrioVoteCounts
            Label decision counts for the aligned trio of classifiers.

        Returns
        -------
        None.

        """
        self.vote_counts = vote_counts
        self.vote_frequencies = self.vote_counts.to_frequencies_exact()
        self.labels = ("a", "b")

        self.majority_right_vote_patterns = {
            label: [
                votes
                for votes in self.vote_patterns
                if votes.count(label) >= 2
            ]
            for label in self.labels
        }

        self.majority_wrong_vote_patterns = {
            label: [
                votes
                for votes in self.vote_patterns
                if votes.count(label) <= 1
            ]
            for label in self.labels
        }

        self.evaluation_exact = [
            self.compute_vote_pattern_evaluation(vps, flip)
            for vps, flip in [
                (self.majority_right_vote_patterns, False),
                (self.majority_wrong_vote_patterns, False),
            ]
        ]

        self.evaluation_float = [
            self.to_float(sol) for sol in self.evaluation_exact
        ]

    def compute_vote_pattern_evaluation(self, vote_patterns, flip):
        return {
            "prevalence": self.prevalences(vote_patterns),
            "accuracy": [
                {
                    label: self.classifier_label_accuracy(
                        classifier, vote_patterns, label, flip
                    )
                    for label in self.labels
                }
                for classifier in range(3)
            ],
        }

    def prevalences(self, vote_patterns):
        """Compute label prevalences in the test."""
        return {
            label: sum(
                [self.vote_frequencies[vp] for vp in vote_patterns[label]]
            )
            for label in self.labels
        }

    def classifier_label_accuracy(
        self, classifier, vote_patterns, label, flip
    ):
        """Compute the label accuracy for classifier."""
        if flip:
            if label == "a":
                correct_label = "b"
            elif label == "b":
                correct_label = "a"
        else:
            correct_label = label
        return sum(
            [
                self.vote_frequencies[vp]
                for vp in vote_patterns[correct_label]
                if vp[classifier] == correct_label
            ]
        ) / sum(
            [self.vote_frequencies[vp] for vp in vote_patterns[correct_label]]
        )

    def to_float(self, sol):
        as_floats = {}
        as_floats["prevalence"] = {
            label: float(val) for label, val in sol["prevalence"].items()
        }
        as_floats["accuracy"] = [
            ({label: float(val) for label, val in classifier_dict.items()})
            for classifier_dict in sol["accuracy"]
        ]
        return as_floats


if __name__ == "__main__":
    from pprint import pprint

    sympy.init_printing(use_unicode=True)

    print(
        """The evaluation observed voting patterns by true label -
        the ground truth."""
    )
    pprint(uciadult_label_counts)

    print()

    print(
        """During unlabeled evaluation we only get to observe the voting
        patterns, without knowing the true label of each voting instance."""
    )
    data_sketch = TrioLabelVoteCounts(
        uciadult_label_counts
    ).to_TrioVoteCounts()

    print(
        """To carry out the evaluation we need the relative frequency of the
        voting patterns."""
    )
    observed_frequencies = data_sketch.to_frequencies_exact()
    pprint(observed_frequencies)

    print()

    print(
        """To estimate the prevalence of the alpha label, we need the
    coefficients of the prevalence quadratic:"""
    )
    error_ind_evaluator = ErrorIndependentEvaluation(data_sketch)
    prev_terms = error_ind_evaluator.alpha_prevalence_quadratic_terms()

    print("1. The 'a' coefficient for the quadratic:")
    pprint(prev_terms[2])
    pprint("2. The 'b' coefficient for the quadratic:")
    pprint(prev_terms[1])
    pprint("3. The 'c' coefficient for the quadratic:")
    pprint(prev_terms[0])

    print("Algebraic evaluation in its exact form: ")
    pprint(error_ind_evaluator.evaluation_exact, sort_dicts=False)
    print()
    print("Algebraic evaluation with inexact floats: ")
    pprint(error_ind_evaluator.evaluation_float, sort_dicts=False)

    gt = SupervisedEvaluation(TrioLabelVoteCounts(uciadult_label_counts))
    print()

    trueAPrevalence = gt.evaluation_exact["prevalence"]["a"]
    print("The true alpha label prevalence is: ")
    pprint(trueAPrevalence)
    print(" or: ")
    pprint(float(trueAPrevalence))

    print()

    # The test run picked for this code comes from a trio of classifiers
    # that have, in fact, very small pair error correlations.
    print("Ground truth values:")
    pprint(gt.evaluation_exact, sort_dicts=False)
    print("Ground truth as inexact floats:")
    pprint(gt.evaluation_float, sort_dicts=False)

    print("The majority voting evaluator:")
    mv_eval = MajorityVotingEvaluation(data_sketch)
    pprint(mv_eval.evaluation_exact, sort_dicts=False)
    print("As floats:")
    pprint(mv_eval.evaluation_float, sort_dicts=False)



"""Axioms of evaluation for R=3 tests.

The number of decision patterns when 3 responses to a question are possible
is 3^N, where N is the number of responders. As in the R=2 axioms case,
we can start with statistics of correctness in a test that only involve a
single responder. We begin the build out of the R=3 axioms with the N=1
axioms, in other words. Then we will proceed to the N=2 axioms, the
N=3 postulates and so on.

Although we suspect that the N=3, R=3 error independent model also has a
closed algebraic solution, we have failed to get one using 128GB of RAM.
But we are publishing the N=1 case now since it is solvable. In addition,
it is possible to build logical alarms for misaligned classifiers just
using the single classifier axioms. This is good news since the single
classifier case axioms are easily constructed for any finite number of
classes/responses.

Variables
---------

The case of R=3 is interesting because it brings into relief what is barely
visible in the R=2 case - there are many more ways to be wrong than right.
In binary response tests there is only way to be wrong. Now we have two ways
of responding incorrectly.

But the normalization condition for being right and wrong still holds.
The sum of correct answers and incorrect answers must sum to the number
of questions on the test. So we have gone from something like,
    x + y = 1
to
    x + y + z = 1
So what variables should we choose to simplify the math of the generating
polynomials? The "natural" choice is to express the label accuracies in
terms of the two wrong choices. And this is a choice that is going to work
for any R so it is the one we choose here.

So, for example, assuming the three labels are given by (a, b, c). The 'a'
label accuracy would be expressed in terms of the two error rates as,
P_i_a_a = 1 - P_i_b_a - P_i_c_a
"""

import sympy
from ntqr.statistics import SingleClassifierVariables


class SingleClassifierAxioms:
    """
    The axioms of a single R=3 classifier

    ...

    Attributes
    ----------

    question_numbers : List[simpy.Symbol..]
        The variables for the count of correct questions for each true
        label.

    responses : List[simpy.Symbol..]
        The variables for the observed interger count of the three classes
        ('a', 'b', 'c') in the test.

        Rai : number of 'a' responses
        Rbi : number of 'b' responses
        Rci : number of 'c' responses

    correctness_variables: List[simpy.Symbol..]
        The variables associated with correct and wrong responses given
        the true label.

        The variables associated with true label 'a'
        Raia : number of 'a' responses (also number of correct)
        Rbia : number of 'b' responses
        Rcia : number of 'c' responses

        The variables associated with true label 'b'
        Raib : number of 'a' responses
        Rbib : number of 'b' responses  (also number of correct)
        Rcib : number of 'c' responses

        The variables associated with true label 'c'
        Raic : number of 'a' responses
        Rbic : number of 'b' responses
        Rcic : number of 'c' responses (also number of correct)


    Methods
    -------
    evaluate_axioms(eval_dict): List[expression, expression, expression]
        Evaluates the axioms given the variable substitutions in 'eval_dict'.
    satisfies_axioms(eval_dict): Boolean
        Checks if the variable substitutions in 'eval_dict' make all three
        axioms identically zero.
    """

    def __init__(self, labels, classifier):
        "Constructs variables for 'labels' and the axioms they must satisfy."

        self.labels = labels
        self.classifier = classifier

        vars = SingleClassifierVariables(labels, classifier)
        self.questions_number = vars.questions_number
        self.responses = vars.responses
        self.responses_by_label = vars.responses_by_label

        # Construct the three, dependent axioms for a single classifier
        self.algebraic_expressions = {}
        for i_true in range(len(labels)):

            true_label = labels[i_true]
            true_label_responses = self.responses_by_label[true_label]
            q_number = self.questions_number[true_label]

            mistakes_out_of_label = q_number
            mistakes_into_label = 0
            for i_response in range(len(labels)):
                response_label = labels[i_response]
                if i_response != i_true:
                    mistakes_out_of_label -= true_label_responses[
                        response_label
                    ]
                    mistakes_into_label += self.responses_by_label[
                        response_label
                    ][true_label]

            self.algebraic_expressions[true_label] = (
                mistakes_out_of_label
                + mistakes_into_label
                - self.responses[true_label]
            )

    def evaluate_axioms(self, eval_dict):
        """
        Evaluates axioms given 'eval_dict'.

        Parameters
        ----------
        eval_dict: Map[sympySymbol -> value]

        Returns
        -------
        Dict mapping label to axiom expression.
        """

        return {
            label: axiom.subs(eval_dict)
            for label, axiom in self.algebraic_expressions.items()
        }

    def satisfies_axioms(self, eval_dict):
        """
        Tests axioms are satisfied for 'eval_dict' values.

        Parameters
        ----------
        eval_dict: Map[sympySymbol -> value]

        Returns
        -------
        Boolean: returns True if the axioms are identically zero, False
        otherwise.
        """

        evaluated_axioms = self.evaluate_axioms(eval_dict)
        return all([(axiom == 0) for axiom in evaluated_axioms.values()])



"""
Module with classes that encapsulate the production of statistical variables
for evaluation models.
"""

import sympy


class SingleClassifierVariables:
    """Statistical variables associated with a single classifier.

    Attributes
    ----------

    question_numbers : List[simpy.Symbol..]
        The variables for the count of correct questions for each true
        label.

    responses : List[simpy.Symbol..]
        The variables for the observed interger count of labels in
        the test. Generally, of the form

        R_{l, i} : Number of 'l' label responses by classifier 'i'

    correctness_variables: List[simpy.Symbol..]
        The variables associated with correct and wrong responses given
        the true label.

        R_{l_r_{i}, l_true} : Number of l_r responses by classifier 'i'
        given true label, l_true.
    """

    def __init__(self, labels, classifier):
        "Constructs variables for 'labels'."

        self.questions_number = {
            label: sympy.Symbol(r"Q_" + label) for label in labels
        }
        self.responses = self.response_variables(labels, classifier)
        self.responses_by_label = self.label_response_variables(
            labels, classifier
        )

    def response_variables(self, labels, classifier):
        """
        Constructs observable response variables given 'labels' and
        'classifier'.

        Parameters
        ----------
        labels : List
            Labels to use.
        classifier : int
            Index of classifier.

        Returns
        -------
        Dictionary of by-label response counts, one per label.
        """

        clsfr_str = str(classifier)
        vars = {}
        for label in labels:
            vars[label] = sympy.Symbol(
                r"R_{" + label + r"_{" + clsfr_str + r"}}"
            )

        return vars

    def label_response_variables(self, labels, classifier):
        """
        Constructs variables associated with correct and wrong
        response counts given true label.

        Parameters
        ----------
        labels : List
            Labels to use.
        classifier : int
            Index of classifier.

        Returns
        -------
        Dictionary of by-label response counts, three per label.
        In addition, each label contains a 'correct' dictionary
        that contains the true_label and its associated count.
        An 'errors' dictionary is indexed by possible wrong
        label assignments.
        """

        clsfr_str = str(classifier)
        vars = {}
        for true_label in labels:
            label_vars = vars.setdefault(true_label, {})
            label_vars_correct = label_vars.setdefault("correct", {})
            label_vars_error = label_vars.setdefault("errors", {})

            for response_label in labels:
                curr_var = sympy.Symbol(
                    r"R_{"
                    + response_label
                    + r"_{"
                    + clsfr_str
                    + r"},"
                    + true_label
                    + r"}"
                )
                label_vars[response_label] = curr_var

                # Now we add it to either 'correct' or 'errors'
                if response_label != true_label:
                    label_vars_error[response_label] = curr_var
                else:
                    label_vars_correct[true_label] = curr_var

        return vars



"""Tests related to the ntqr.r3.raxioms module."""

import ntqr

r3axioms = ntqr.r3.raxioms.SingleClassifierAxioms(["a", "b", "c"], "k")

mul = 2
qa = 2 * mul
qb = 3 * mul
qc = 4 * mul

# Setting the question numbers
eval_dict = {
    qlabel: val
    for qlabel, val in zip(r3axioms.questions_number.values(), [qa, qb, qc])
}

# Setting the observed label responses
eval_dict.update(
    {
        response_var: count
        for response_var, count in zip(
            r3axioms.responses.values(), [1 * mul, 2 * mul, 6 * mul]
        )
    }
)


def test_r3_single_classifier_axioms():
    rbl_vars = r3axioms.responses_by_label
    eval_dict.update(
        {
            rbl_vars["a"]["b"]: 1,
            rbl_vars["a"]["c"]: 2,
            rbl_vars["b"]["a"]: 0,
            rbl_vars["b"]["c"]: 5,
            rbl_vars["c"]["a"]: 1,
            rbl_vars["c"]["b"]: 2,
        }
    )
    evaluated_axioms = r3axioms.evaluate_axioms(eval_dict)
    assert evaluated_axioms == {"a": 0, "b": 0, "c": 0}



import ntqr.alarms
import ntqr.r3.raxioms
import ntqr.r3.evaluations

import pytest

labels = ("a", "b", "c")
c1Axioms = ntqr.r3.raxioms.SingleClassifierAxioms(labels, "1")
c2Axioms = ntqr.r3.raxioms.SingleClassifierAxioms(labels, "2")
alarm = ntqr.alarms.SingleClassifierAxiomAlarm(
    18, [c1Axioms, c2Axioms], ntqr.r3.evaluations.SingleClassifierEvaluations
)

# Better than 50%
factors = (2, 2, 2)
alarm.set_safety_specification(factors)
qs = (4, 6, 8)


@pytest.mark.parametrize(
    "qs, factors, corrects, satisfies",
    (
        (qs, factors, (3, 4, 5), True),
        (qs, factors, (2, 4, 5), False),
        (qs, factors, (3, 3, 5), False),
    ),
)
def test_safety_specification(qs, factors, corrects, satisfies):
    safety_specification = ntqr.alarms.LabelsSafetySpecification(factors)
    assert safety_specification.is_satisfied(qs, corrects) == satisfies


@pytest.mark.parametrize(
    "qs, responses, misaligned",
    (
        (qs, ((4, 7, 7), (3, 7, 8)), False),
        (qs, ((10, 8, 0), (0, 8, 10)), True),
    ),
)
def test_misalignment_alarm_at_qs(qs, responses, misaligned):

    assert alarm.misaligned_at_qs(qs, responses) == misaligned



"""
Evaluations for binary tests (R=2).

For any finite test there is a finite set of evaluations possible. The
classes in this module compute them.

Classes:
    SingleClassifierEvaluations: Class related to the evaluations for a
    single classifier consistent with its observed test responses.

Functions:

Misc variables:

"""

import itertools

import sympy

import ntqr.evaluations
from ntqr.r2.datasketches import VoteCounts


class SingleClassifierEvaluations(
    ntqr.evaluations.SingleClassifierEvaluations
):
    """
    Single classifier evaluations in (Q_a, Q_b, R_{b_i, a}, R_{a_i,b})
    space.
    """

    def __init__(self, Q, single_axioms):

        super().__init__(Q, single_axioms)

    def number_apriori_evaluations(self):
        """
        Calculates all the possible evaluations for a binary response test
        with Q questions.

        Returns
        -------
        int

        """
        q = self.Q
        return 1 / 6 * (q + 1) * (q + 2) * (q + 3)

    def errors_at_qs(self, qs, responses):
        """
        Returns all evaluations logically consistent with the
        single classifier axiom given the correct number of each
        label and a classifier's responses.

        In binary classification we have Q_a + Q_b = Q. Thus, we
        really need to specify only two of the three (Q, Q_a, Q_b).
        Making a choice is arbitrary and breaks the symmetry in the
        algebra between the two labels. Instead, we specify Q_a, and
        Q_b and since we have Q from the instance initialization, we
        do a quality check (logic check) of the equality between the
        three quantities.
        """
        eval_dict = {
            var: val
            for var, val in zip(self.axioms.questions_number.values(), qs)
        }
        eval_dict.update(
            {
                var: val
                for var, val in zip(self.axioms.responses.values(), responses)
            }
        )

        wrong_vars = [
            [
                wrong_var
                for wrong_var in self.axioms.responses_by_label[true_label][
                    "errors"
                ].values()
            ]
            for true_label in self.axioms.labels
        ]

        evals = set(
            (first_lbl_wrong, second_lbl_wrong)
            for first_lbl_wrong in self._label_wrongs_(qs[0])
            for second_lbl_wrong in self._label_wrongs_(qs[1])
            if self._check_axiom_consistency_(
                eval_dict,
                itertools.chain(*wrong_vars),
                itertools.chain(first_lbl_wrong, second_lbl_wrong),
            )
        )

        return evals

    def max_correct_at_qs(self, qs, responses):
        """Gives highest performing correct for each label.

        Meant to save memory for alarm applications.
        """

        eval_dict = {
            var: val
            for var, val in zip(self.axioms.questions_number.values(), qs)
        }
        eval_dict.update(
            {
                var: val
                for var, val in zip(self.axioms.responses.values(), responses)
            }
        )

        wrong_vars = [
            [
                wrong_var
                for wrong_var in self.axioms.responses_by_label[true_label][
                    "errors"
                ].values()
            ]
            for true_label in self.axioms.labels
        ]

        max_correct = (0, (0, 0))

        for first_label_wrongs in self._label_wrongs_(qs[0]):
            vars = wrong_vars[0]
            vals = first_label_wrongs
            for second_label_wrongs in self._label_wrongs_(qs[1]):
                vars += wrong_vars[1]
                vals += second_label_wrongs

                if self._check_axiom_consistency_(eval_dict, vars, vals):
                    corrects = [
                        ql - sum(wrongs)
                        for ql, wrongs in zip(
                            qs,
                            [first_label_wrongs, second_label_wrongs],
                        )
                    ]
                    corrects_sum = sum(corrects)
                    if corrects_sum > max_correct[0]:
                        max_correct = (corrects_sum, corrects)

        return max_correct[1]

    def all_qs(self):
        """Return all possible question numbers."""
        Q = self.Q
        return [(qa, Q - qa) for qa in range(0, self.Q + 1)]

    def _label_wrongs_(self, ql):
        """Return all possible incorrect given Q_l."""
        return [(num_wrong,) for num_wrong in range(0, ql + 1)]


# This class needs to be deleted and code that uses refactored
class PosteriorSingleEvaluations:
    """Evaluations logically consistent with observed responses."""

    def __init__(self, tvc: VoteCounts):
        self.Q = tvc.test_size
        self.tvc = tvc

    def all_possible_evaluations(self, classifier: int):
        # The b label frequency for the classifier
        Rb = int(self.tvc.classifier_label_frequency(classifier, "b") * self.Q)
        evaluations = {}
        for Qa in range(0, self.Q + 1):
            evaluations[Qa] = []
            for Raa in range(0, Qa + 1):
                for Rbb in range(0, self.Q - Qa + 1):
                    if Raa - Rbb - Qa + Rb == 0:
                        evaluations[Qa].append((Raa, Rbb))

        return evaluations

    def find_k_nearest_at_prevalence(
        self, classifier: int, qa: int, point, k: int
    ):
        # The b label frequency for the classifier
        Rb = int(self.tvc.classifier_label_frequency(classifier, "b") * self.Q)

        Qa = qa
        qa_candidates = []
        for Raa in range(0, Qa + 1):
            for Rbb in range(0, self.Q - Qa + 1):
                if Raa - Rbb - Qa + Rb == 0:
                    distances = self.distances_to_target(Qa, Raa, Rbb, point)
                    qa_candidates.append((distances, (Raa, Rbb)))
        # Since all list items have the same distance to Qa,
        # this sorting returns min total distance
        qa_candidates.sort(key=lambda x: x[0][0] + x[0][1])
        evaluations = qa_candidates[:k]

        return evaluations

    def find_k_nearest_at_prevalence_all_classifiers(
        self, qa: int, points, k: int
    ):
        single_evals = [
            self.find_k_nearest_at_prevalence(
                classifier, qa, points[classifier], k
            )
            for classifier in range(3)
        ]
        joint_evals = []
        for distances0, r0 in single_evals[0]:
            qa_distance = distances0[0]
            r0distance = distances0[1]
            for distances1, r1 in single_evals[1]:
                r1distance = distances1[1]
                for distances2, r2 in single_evals[2]:
                    r2distance = distances2[1]
                    joint_evals.append(
                        (
                            qa_distance + r0distance + r1distance + r2distance,
                            (r0, r1, r2),
                        )
                    )
        joint_evals.sort()
        return joint_evals[:k]

    def distances_to_target(self, Qa, Raa, Rbb, point):
        # point_pspace = self.to_pspace(point)
        pe_pspace = self.to_pspace((Qa, Raa, Rbb))

        return (
            sympy.simplify((point[0] - pe_pspace[0]) ** 2),
            sympy.simplify(
                (point[1] - pe_pspace[1]) ** 2 + (point[2] - pe_pspace[2]) ** 2
            ),
        )

    def to_pspace(self, point):
        if point[0] != 0:
            pia = sympy.Rational(point[1], point[0])
        else:
            pia = sympy.Rational(0, 1)

        if self.Q - point[0] != 0:
            pib = sympy.Rational(point[2], (self.Q - point[0]))
        else:
            pib = sympy.Rational(0, 1)

        return (sympy.Rational(point[0], self.Q), pia, pib)



"""@author: Andrés Corrada-Emmanuel."""
import itertools
from fractions import Fraction

import pytest

from ntqr.r2.datasketches import (
    TrioVoteCounts,
    TrioLabelVoteCounts,
    classifier_label_votes,
)
from ntqr.r2.examples import uciadult_label_counts

trio_vote_patterns = list(itertools.product(*["ab" for i in range(3)]))


def test_classifier_a_label_votes():
    vote_patterns = classifier_label_votes(0, "a", trio_vote_patterns)
    assert vote_patterns == (
        ("a", "a", "a"),
        ("a", "a", "b"),
        ("a", "b", "a"),
        ("a", "b", "b"),
    )


def test_classifier_b_label_votes():
    vote_patterns = classifier_label_votes(0, "b", trio_vote_patterns)
    assert vote_patterns == (
        ("b", "a", "a"),
        ("b", "a", "b"),
        ("b", "b", "a"),
        ("b", "b", "b"),
    )


#
# Testing TrioLabelVoteCounts functionality
#


def test_triolabelvotecounts_init():
    tlvc = TrioLabelVoteCounts(uciadult_label_counts)

    assert isinstance(tlvc, TrioLabelVoteCounts)


#
# Testing TrioVoteCounts functionality
#


def test_triovotecounts_init():
    tlvc = TrioLabelVoteCounts(uciadult_label_counts)
    tvc = tlvc.to_TrioVoteCounts()

    assert isinstance(tvc, TrioVoteCounts)


tvc = TrioLabelVoteCounts(uciadult_label_counts).to_TrioVoteCounts()
uci_freqs = tvc.to_frequencies_exact()


@pytest.mark.parametrize(
    "freqs, votes,freq",
    (
        (uci_freqs, ("a", "a", "a"), Fraction(493, 18421)),
        (uci_freqs, ("a", "b", "a"), Fraction(5801, 36842)),
    ),
)
def test_frequency(freqs, votes, freq):
    assert freqs[votes] == freq


pair_moments = tvc.label_pairs_frequency_moments("b")


@pytest.mark.parametrize(
    "pair_moments, pair, freq",
    (
        (pair_moments, (0, 1), Fraction(18432653, 1357332964)),
        (pair_moments, (0, 2), Fraction(18272925, 1357332964)),
        (pair_moments, (1, 2), Fraction(16803485, 1357332964)),
    ),
)
def test_pair_moment(pair_moments, pair, freq):
    assert pair_moments[pair] == freq



"""Tests to check the ntqr.r2.raxioms module."""

import pytest

import ntqr.r2.raxioms

labels = ["a", "b"]
classifier = "k"


r2axioms = ntqr.r2.raxioms.SingleClassifierAxioms(labels, classifier)

eval_dict = {
    var: val for var, val in zip(r2axioms.questions_number.values(), [3, 7])
}
eval_dict.update(
    {var: val for var, val in zip(r2axioms.responses.values(), [5, 5])}
)

wrong_vars = [
    var
    for true_label in r2axioms.labels
    for var in r2axioms.responses_by_label[true_label]["errors"].values()
]


@pytest.mark.parametrize(
    "update_dict, satisfies_axioms",
    (
        ({var: val for var, val in zip(wrong_vars, (0, 2))}, True),
        ({var: val for var, val in zip(wrong_vars, (1, 2))}, False),
    ),
)
def test_evaluations_at_qa_qb(update_dict, satisfies_axioms):
    eval_dict.update(update_dict)
    test_val = r2axioms.satisfies_axioms(eval_dict)
    assert test_val == satisfies_axioms



"""Module for R-axioms related super classess."""


class SingleClassifierEvaluations:
    """
    Single classifier evaluations
    """

    def __init__(self, Q, single_axioms):

        self.Q = Q
        self.axioms = single_axioms

    def correct_at_qs(self, qs, responses):
        errors_at_qs = self.errors_at_qs(qs, responses)
        return set(
            ((ql - sum(label_errors)) for ql, label_errors in zip(qs, errors))
            for errors in errors_at_qs
        )

    def max_correct_at_qs(self, qs, responses):
        """Gives highest performing correct for each label.

        Meant to save memory for alarm applications.
        """

        eval_dict = {
            var: val
            for var, val in zip(self.axioms.questions_number.values(), qs)
        }
        eval_dict.update(
            {
                var: val
                for var, val in zip(self.axioms.responses.values(), responses)
            }
        )

        wrong_vars = [
            [
                wrong_var
                for wrong_var in self.axioms.responses_by_label[true_label][
                    "errors"
                ].values()
            ]
            for true_label in self.axioms.labels
        ]

        max_correct = (0, (0, 0))

        for first_label_wrongs in self._label_wrongs_(qs[0]):
            vars = wrong_vars[0]
            vals = first_label_wrongs
            for second_label_wrongs in self._label_wrongs_(qs[1]):
                vars += wrong_vars[1]
                vals += second_label_wrongs

                if self._check_axiom_consistency_(eval_dict, vars, vals):
                    corrects = [
                        ql - sum(wrongs)
                        for ql, wrongs in zip(
                            qs,
                            [first_label_wrongs, second_label_wrongs],
                        )
                    ]
                    corrects_sum = sum(corrects)
                    if corrects_sum > max_correct[0]:
                        max_correct = (corrects_sum, corrects)

        return max_correct[1]

    def all_qs(self):
        """Return all possible question numbers."""
        Q = self.Q
        return [(qa, Q - qa) for qa in range(0, self.Q + 1)]

    def _label_wrongs_(self, ql):
        """Return all possible incorrect given Q_l."""
        return [(num_wrong,) for num_wrong in range(0, ql + 1)]

    def _check_axiom_consistency_(self, eval_dict, wrong_vars, wrong_vals):
        eval_dict.update(
            {var: val for var, val in zip(wrong_vars, wrong_vals)}
        )
        return self.axioms.satisfies_axioms(eval_dict)



"""
Evaluate noisy decision makers using logic and algebra.

Classes:

Functions:

Misc variables:

    __version__
    uci_adult_test_example
"""

__version__ = "0.4.2.2"

from ntqr.r2.examples import uciadult_label_counts

from ntqr.r2.datasketches import TrioLabelVoteCounts, TrioVoteCounts

from ntqr.r2.evaluators import (
    SupervisedEvaluation,
    ErrorIndependentEvaluation,
    MajorityVotingEvaluation,
)

import ntqr.statistics
import ntqr.r2
import ntqr.r3



# -*- coding: utf-8 -*-
import sympy

from ntqr.r2.evaluators import (
    ErrorIndependentEvaluation,
    SupervisedEvaluation,
    MajorityVotingEvaluation,
)
from ntqr.r2.datasketches import TrioLabelVoteCounts
from ntqr.r2.examples import uciadult_label_counts


# Supervised evaluation tests


def test_supervisedevaluation():
    tlvc = TrioLabelVoteCounts(uciadult_label_counts)
    se = SupervisedEvaluation(tlvc)

    assert isinstance(se, SupervisedEvaluation)


def test_supervised_prevalences():
    supervised_eval = SupervisedEvaluation(tlvc)

    assert supervised_eval.evaluation_exact == {
        "prevalence": {
            "a": sympy.Rational(5687, 36842),
            "b": sympy.Rational(31155, 36842),
        },
        "accuracy": [
            {
                "a": sympy.Rational(3737, 5687),
                "b": sympy.Rational(6501, 10385),
            },
            {
                "a": sympy.Rational(1260, 5687),
                "b": sympy.Rational(29744, 31155),
            },
            {"a": sympy.Rational(4746, 5687), "b": sympy.Rational(4168, 6231)},
        ],
        "pair_correlation": {
            (0, 1): {
                "a": sympy.Rational(273192, 32341969),
                "b": sympy.Rational(2204576, 323544675),
            },
            (0, 2): {
                "a": sympy.Rational(13325, 32341969),
                "b": sympy.Rational(-79682, 12941787),
            },
            (1, 2): {
                "a": sympy.Rational(-264525, 32341969),
                "b": sympy.Rational(94508, 38825361),
            },
        },
        "3_way_correlation": {
            (0, 1, 2): {
                "a": sympy.Rational(452568508, 183928777703),
                "b": sympy.Rational(-27265589, 134400457995),
            }
        },
    }


tlvc = TrioLabelVoteCounts(uciadult_label_counts)
tvc = tlvc.to_TrioVoteCounts()


# Error independent evaluation tests


def test_errorindependentevaluation():
    eieval = ErrorIndependentEvaluation(tvc)

    assert isinstance(eieval, ErrorIndependentEvaluation)


# Majority voting evaluation tests


def test_majorityevaluation():
    mv_eval = MajorityVotingEvaluation(tvc)

    assert isinstance(mv_eval, MajorityVotingEvaluation)


def test_mv_evaluation():
    mv_eval = MajorityVotingEvaluation(tvc)

    assert mv_eval.evaluation_exact == [
        {
            "prevalence": {
                "a": sympy.Rational(7979, 36842),
                "b": sympy.Rational(28863, 36842),
            },
            "accuracy": [
                {
                    "a": sympy.Rational(7417, 7979),
                    "b": sympy.Rational(20891, 28863),
                },
                {
                    "a": sympy.Rational(2178, 7979),
                    "b": sympy.Rational(28370, 28863),
                },
                {
                    "a": sympy.Rational(7349, 7979),
                    "b": sympy.Rational(21151, 28863),
                },
            ],
        },
        {
            "prevalence": {
                "a": sympy.Rational(28863, 36842),
                "b": sympy.Rational(7979, 36842),
            },
            "accuracy": [
                {
                    "a": sympy.Rational(7972, 28863),
                    "b": sympy.Rational(562, 7979),
                },
                {
                    "a": sympy.Rational(493, 28863),
                    "b": sympy.Rational(5801, 7979),
                },
                {
                    "a": sympy.Rational(7712, 28863),
                    "b": sympy.Rational(630, 7979),
                },
            ],
        },
    ]



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
from pathlib import Path
from ntqr import __version__ as ntqr_version


sys.path.insert(0, str(Path("..", "..", "src").resolve()))
# sys.path.insert(0, str(Path("..", "src").resolve()))
# sys.path.insert(0, str(Path("..").resolve()))


# -- Project information -----------------------------------------------------

project = "NTQR"
copyright = "2024, Andrés Corrada-Emmanuel, Walker Lee, Adam Sloat"
authors = ["Andrés Corrada-Emmanuel", "Walker Lee", "Adam Sloat"]

# The full version, including alpha/beta/rc tags
version = ntqr_version
release = version


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "myst_parser",
    "sphinx.ext.mathjax",
    "nbsphinx",
    "autoapi.extension",
]

autoapi_dirs = [str(Path("..", "..", "src"))]
autoapi_type = "python"
autoapi_generate_api_docs = True
# autoapi_modules = {"ntqr": None}
autoapi_keep_files = False
autoapi_template_dir = "_templates/autoapi/python"
# autosummary_generate = True  # Turn on sphinx.ext.autosummary
# autodoc_mock_imports = ["ntqr"]

myst_enable_extensions = [
    "colon_fence",
    "dollarmath",
    "amsmath",
]

mathjax_path = "https://cdn.jsdelivr.net/npm/mathjax@2/MathJax.js?config=TeX-AMS-MML_HTMLorMML"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "furo"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]



# -*- coding: utf-8 -*-
"""R=2 evaluation axioms in percentage space.

This is the space of rationals - ratios of integers - where the
sample statistics of correctness in a test are expressed as sample
percentages.
"""

import sympy

# The prevalences of the two labels
pa = sympy.Symbol(r"P_a")
pb = sympy.Symbol(r"P_b")

# The label accuracy for the classifiers "i", "j", "k"
pia = sympy.Symbol(r"P_{i, a}")
pib = sympy.Symbol(r"P_{i, b}")
pja = sympy.Symbol(r"P_{j, a}")
pjb = sympy.Symbol(r"P_{j, b}")
pka = sympy.Symbol(r"P_{k, a}")
pkb = sympy.Symbol(r"P_{k, b}")

# Axioms before observing test results
# One postulate is just about the prevalences
prelances_sum_to_one = pa + pb - 1

# We omit all the linear relations that relate binary erorrs to accuracy.
# By this is meant that we could work in a larger variable space that
# included the error rate for a label. This would require that we
# add equations of the type P_i_a_a + P_i_b_a = 1 to the generating set.
# So instead we just express everything in therms of binary label accuracy
# and dispense with writing out error rate variables.
# Dropping the error rate variables also allows the syntactic sugar of
# referring to P_i_a_a (the percentage of times classifier i said the label
# was "a" AND it was "a")as just P_i_a.

# Postulates having observed responses
fai = sympy.Symbol(r"f_{a_i}")
fbi = sympy.Symbol(r"f_{b_i}")
faj = sympy.Symbol(r"f_{a_j}")
fbj = sympy.Symbol(r"f_{b_j}")
fak = sympy.Symbol(r"f_{a_k}")
fbk = sympy.Symbol(r"f_{b_k}")

# The 'generating set' of polynomials for any binary classifier
single_binary_classifier_generating_set = [
    pa * pia + pb * (1 - pib) - fai,
    pa * (1 - pia) + pb * pib - fbi,
]

# The single classifier generating set only creates one postulate
single_binary_classifier_axiom = pa * (pia - fai) - pb * (pib - fbi)

# Data sketch observables for the pair
faiaj = sympy.Symbol(r"f_{a_i, a_j}")
faibj = sympy.Symbol(r"f_{a_i, b_j}")
fbiaj = sympy.Symbol(r"f_{b_i, a_j}")
fbibj = sympy.Symbol(r"f_{b_i, b_j}")

# The delta moment for classifier pairs
deltaij = sympy.Symbol(r"\Delta_{i,j}")

# Pair correlation variables for the two labels
gija = sympy.Symbol(r"\Gamma_{i, j, a}")
gijb = sympy.Symbol(r"\Gamma_{i, j, b}")
gika = sympy.Symbol(r"\Gamma_{i, k, a}")
gikb = sympy.Symbol(r"\Gamma_{i, k, b}")
gjka = sympy.Symbol(r"\Gamma_{j, k, a}")
gjkb = sympy.Symbol(r"\Gamma_{j, k, b}")

# 3-way correlations for the two labels
gijka = sympy.Symbol(r"\Gamma_{i, j, k, a}")
gijkb = sympy.Symbol(r"\Gamma_{i, j, k, b}")

# The 'generating set' of polynomials for any pair of binary classifiers
pair_binary_classifiers_generating_set = [
    pa * (pia * pja + gija) + pb * ((1 - pib) * (1 - pjb) + gijb) - faiaj,
    pa * (pia * (1 - pja) - gija) + pb * ((1 - pib) * pjb - gijb) - faibj,
    pa * ((1 - pia) * pja - gija) + pb * (pib * (1 - pjb) - gijb) - fbiaj,
    pa * ((1 - pia) * (1 - pja) + gija) + pb * (pib * pjb + gijb) - fbibj,
]

# The Groebner basis for the classifier pair generating set produces a
# set of partially independent postulates
pair_binary_classifiers_axioms = [
    (pia - fai) * (pjb - fbj) - (pib - fbi) * (pja - faj),
    (pia - fai) * (pjb - fbj) * ((pia - fai) + (pib - fbi))
    + (pia - fai) * (gijb - deltaij)
    + (pib - fbi) * (gija - deltaij),
    (pia - fai) * (pjb - fbj) * ((pja - faj) + (pjb - fbj))
    + (pja - faj) * (gijb - deltaij)
    + (pjb - fbj) * (gija - deltaij),
]

# Data sketch observables for the trio
faiajak = sympy.Symbol(r"f_{a_i, a_j, a_k}")
faiajbk = sympy.Symbol(r"f_{a_i, a_j, b_k}")
faibjak = sympy.Symbol(r"f_{a_i, b_j, a_k}")
fbiajak = sympy.Symbol(r"f_{b_i, a_j, a_k}")

fbibjbk = sympy.Symbol(r"f_{b_i, b_j, b_k}")
fbibjak = sympy.Symbol(r"f_{b_i, b_j, a_k}")
fbiajbk = sympy.Symbol(r"f_{b_i, a_j, b_k}")
faibjbk = sympy.Symbol(r"f_{a_i, b_j, b_k}")

trio_frequencies = [
    faiajak,
    faiajbk,
    faibjak,
    faibjbk,
    fbiajak,
    fbiajbk,
    fbibjak,
    fbibjbk,
]

trio_binary_classifiers_generating_set = [
    pa * (pia * pja * pka + gija * pka + gika * pja + gjka * pia + gijka)
    + pb
    * (
        (1 - pib) * (1 - pjb) * (1 - pkb)
        + gijb * (1 - pkb)
        + gikb * (1 - pjb)
        + gjkb * (1 - pib)
        - gijkb
    )
    - faiajak,
    pa
    * (
        -gijka
        + (1 - pka) * gija
        - pja * gika
        - pia * gjka
        + pia * pja * (1 - pka)
    )
    + pb
    * (
        gijkb
        + pkb * gijb
        - (1 - pjb) * gikb
        - (1 - pib) * gjkb
        + (1 - pib) * (1 - pjb) * pkb
    )
    - faiajbk,
    pa
    * (
        pia * (1 - pja) * pka
        - pka * gija
        + (1 - pja) * gika
        - pia * gjka
        - gijka
    )
    + pb
    * (
        (1 - pib) * pjb * (1 - pkb)
        - (1 - pkb) * gijb
        + pjb * gikb
        - (1 - pib) * gjkb
        + gijkb
    )
    - faibjak,
    pa
    * (
        pia * (1 - pja) * (1 - pka)
        - (1 - pka) * gija
        - (1 - pja) * gika
        + pia * gjka
        + gijka
    )
    + pb
    * (
        (1 - pib) * pjb * pkb
        - pkb * gijb
        - pjb * gikb
        + (1 - pib) * gjkb
        - gijkb
    )
    - faibjbk,
    pa
    * (
        (1 - pia) * pja * pka
        - pka * gija
        - pja * gika
        + (1 - pia) * gjka
        - gijka
    )
    + pb
    * (
        pib * (1 - pjb) * (1 - pkb)
        - (1 - pkb) * gijb
        - (1 - pjb) * gikb
        + pib * gjkb
        + gijkb
    )
    - fbiajak,
    pa
    * (
        (1 - pia) * pja * (1 - pka)
        - (1 - pka) * gija
        + pja * gija
        - (1 - pia) * gjka
        + gijka
    )
    + pb
    * (
        pib * (1 - pjb) * pkb
        - pkb * gijb
        + (1 - pjb) * gikb
        - pib * gjkb
        - gijkb
    )
    - fbiajbk,
    pa
    * (
        (1 - pia) * (1 - pja) * pka
        + pka * gija
        - (1 - pja) * gika
        - (1 - pia) * gjka
        + gijka
    )
    + pb
    * (
        pib * pjb * (1 - pkb)
        + (1 - pkb) * gijb
        - pjb * gikb
        - pib * gjkb
        - gijkb
    )
    - fbibjak,
    pa
    * (
        (1 - pia) * (1 - pja) * (1 - pka)
        + (1 - pka) * gija
        + (1 - pja) * gika
        + (1 - pia) * gjka
        - gijka
    )
    + pb * (pib * pjb * pkb + pkb * gijb + pjb * gikb + pib * gjkb + gijkb)
    - fbibjbk,
]



"""
Plotting functions for binary tests (R=2).

Functions:
    plot_evaluations

Misc variables:

"""
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import matplotlib.patches as mpatches


def prepare_accuracy_axes(axes, title):
    axes.set(title=title, xlim=(0, 100), ylim=(0, 100))

    for dax in (axes.xaxis, axes.yaxis):
        dax.set(
            major_locator=ticker.FixedLocator([50]),
            major_formatter=ticker.PercentFormatter(),
            minor_locator=ticker.FixedLocator(
                [0, 10, 20, 30, 40, 60, 70, 80, 90, 100]
            ),
            minor_formatter=ticker.PercentFormatter(),
        )

    # Styling them
    axes.tick_params(
        axis="both", which="major", grid_color="white", grid_linewidth=2
    )
    axes.tick_params(
        axis="both", which="minor", grid_color="k", grid_alpha=0.4
    )

    # Blindspot line
    axes.plot([0, 100], [100, 0], color="black")

    # The labels for the four quadrants of the unit cube
    quad_label_info = [
        ("A", (75, 75)),
        ("C", (25, 25)),
        ("$F_1$", (25, 75)),
        ("$F_2$", (75, 25)),
    ]
    for quad_label, pos in quad_label_info:
        axes.annotate(
            quad_label,
            pos,
            size=48,
            color="white",
            alpha=0.5,
            textcoords="offset points",
            xytext=(-12, -12),
        )

    # Color the quadrants
    # axes.axhspan(0, 0.5, facecolor="0.5")
    red_color = "#FF8C8C"
    nasa_blue = "#0B3D91"
    light_blue = "#ADD8E6"  # "#E4F6F8"  # "#ADD8E6"
    orange_color = "#F6B26B"
    alpha = 1.0
    axes.axvspan(50, 100, ymin=0.5, ymax=1, color=light_blue, alpha=alpha)
    axes.axvspan(0, 50, ymin=0, ymax=0.5, color=orange_color, alpha=alpha)
    axes.axhspan(50, 100, xmin=0, xmax=0.5, color=red_color, alpha=alpha)
    axes.axhspan(0, 50, xmin=0.5, xmax=1, color=red_color, alpha=alpha)

    axes.set_xlabel(r"$P_{i,a}$", fontsize=12)
    axes.set_ylabel(r"$P_{i,b}$", fontsize=12)

    axes.set_aspect("equal", "box")

    return


def _plot_evaluations(
    ax,
    evals,
    title="Title",
    legend_loc="best",
    withArrows=False,
    classifier_labels=None,
    horizontalalignment="left",
):
    prepare_accuracy_axes(ax, title)

    # Specifying the grid
    ax.grid("on")

    texts_to_adjust = []
    for label, x, y, marker, colors, sizes in evals:
        ax.scatter(
            x,
            y,
            s=sizes,
            color=colors,
            marker=marker,
            label=label,
            alpha=0.9,
            clip_on=False,
        )
        # Adding the classifier label to each point
        for i in range(len(x)):
            if classifier_labels:
                i_label = classifier_labels[i]
            else:
                i_label = str(i + 1)
            texts_to_adjust.append(
                ax.annotate(
                    i_label,
                    (x[i], y[i]),
                    color="red",
                    alpha=1.0,
                    size="18",
                    usetex=False,
                    horizontalalignment=horizontalalignment,
                )
            )

    if withArrows:
        # Since we are comparing evaluations, both evaluations have
        # the same number of points
        x_starts = list(evals[0][1])
        y_starts = list(evals[0][2])
        x_ends = list(evals[1][1])
        y_ends = list(evals[1][2])
        for i in range(len(x_starts)):
            x_start = float(x_starts[i])
            y_start = float(y_starts[i])
            x_end = float(x_ends[i])
            y_end = float(y_ends[i])
            arrow = mpatches.FancyArrowPatch(
                (x_start, y_start),
                (x_end, y_end),
                connectionstyle="angle3",
                arrowstyle="->",
                shrinkA=10,
                shrinkB=10,
                mutation_scale=9,
            )
            ax.add_patch(arrow)

    ax.legend(loc=legend_loc, facecolor="white", labelcolor="black")


def compare_evaluations(
    evals,
    gt_index,
    titles=[],
    figsize=(10, 5),
    legend_loc="best",
    withArrows=False,
    suptitle=None,
    classifier_labels=None,
):
    fig, axs = plt.subplots(
        1, len(evals) - 1, figsize=figsize, layout="constrained"
    )

    j = 0
    for i in range(len(evals)):
        if i == gt_index:
            continue

        _plot_evaluations(
            axs[j],
            (evals[gt_index], evals[i]),
            title=titles[i],
            legend_loc=legend_loc,
            withArrows=withArrows,
            classifier_labels=classifier_labels,
        )
        j += 1

    if suptitle:
        plt.suptitle(suptitle)
    plt.show()

    return


def plot_evaluations(
    evals,
    title="Title",
    figsize=(5, 4),
    legend_loc="lower left",
    withArrows=False,
    classifier_labels=None,
):
    plt.style.use("_mpl-gallery")

    # plot
    fig, ax = plt.subplots(figsize=figsize, layout="constrained")

    _plot_evaluations(
        ax,
        evals,
        title=title,
        legend_loc=legend_loc,
        withArrows=withArrows,
        classifier_labels=classifier_labels,
    )

    plt.show()

    return



"""
Evaluations for binary tests (R=3).

For any finite test there is a finite set of evaluations possible. The
classes in this module compute them for the case of 3 classes or test
responses.

Classes:
    SingleClassifierEvaluations: Class related to the evaluations for a
    single classifier consistent with its observed test responses.

Functions:

Misc variables:

"""

import itertools

import sympy
import ntqr.evaluations


class SingleClassifierEvaluations(
    ntqr.evaluations.SingleClassifierEvaluations
):
    """
    Single classifier evaluations in (Q_a, Q_b, R_{b_i, a}, R_{a_i,b})
    space.
    """

    def __init__(self, Q, single_axioms):

        super().__init__(Q, single_axioms)

    def number_aprior_evaluations(self):
        """
        Calculates all the possible evaluations for a binary response test
        with Q questions.

        Returns
        -------
        int

        """
        Q = self.Q
        prod = 1
        for i in range(1, 10):
            prod *= Q + i
        return prod / 362880  # 9! division

    def errors_at_qs(self, qs, responses):
        """
        Returns all evaluations logically consistent with the
        single classifier axiom given the correct number of each
        label and a classifier's responses.
        """
        eval_dict = {
            var: val
            for var, val in zip(self.axioms.questions_number.values(), qs)
        }
        eval_dict.update(
            {
                var: val
                for var, val in zip(self.axioms.responses.values(), responses)
            }
        )

        wrong_vars = [
            [
                wrong_var
                for wrong_var in self.axioms.responses_by_label[true_label][
                    "errors"
                ].values()
            ]
            for true_label in self.axioms.labels
        ]

        evals = set(
            (first_lbl_wrong, second_lbl_wrong, third_lbl_wrong)
            for first_lbl_wrong in self._label_wrongs_(qs[0])
            for second_lbl_wrong in self._label_wrongs_(qs[1])
            for third_lbl_wrong in self._label_wrongs_(qs[2])
            if self._check_axiom_consistency_(
                eval_dict,
                itertools.chain(*wrong_vars),
                itertools.chain(
                    first_lbl_wrong, second_lbl_wrong, third_lbl_wrong
                ),
            )
        )

        return evals

    def max_correct_at_qs(self, qs, responses):
        """Gives highest performing correct for each label.

        Meant to save memory for alarm applications.
        """

        eval_dict = {
            var: val
            for var, val in zip(self.axioms.questions_number.values(), qs)
        }
        eval_dict.update(
            {
                var: val
                for var, val in zip(self.axioms.responses.values(), responses)
            }
        )

        wrong_vars = [
            [
                wrong_var
                for wrong_var in self.axioms.responses_by_label[true_label][
                    "errors"
                ].values()
            ]
            for true_label in self.axioms.labels
        ]

        max_correct = (0, (0, 0, 0))

        for first_label_wrongs in self._label_wrongs_(qs[0]):
            vars = wrong_vars[0]
            vals = first_label_wrongs
            for second_label_wrongs in self._label_wrongs_(qs[1]):
                vars += wrong_vars[1]
                vals += second_label_wrongs
                for third_label_wrong in self._label_wrongs_(qs[2]):
                    vars += wrong_vars[2]
                    vals += third_label_wrong

                    if self._check_axiom_consistency_(eval_dict, vars, vals):
                        corrects = [
                            ql - sum(wrongs)
                            for ql, wrongs in zip(
                                qs,
                                [
                                    first_label_wrongs,
                                    second_label_wrongs,
                                    third_label_wrong,
                                ],
                            )
                        ]
                        corrects_sum = sum(corrects)
                        if corrects_sum > max_correct[0]:
                            max_correct = (corrects_sum, corrects)

        return max_correct[1]

    def all_qs(self):
        "Returns all possible question numbers."
        Q = self.Q
        return [
            (qa, qb, Q - qa - qb)
            for qa in range(0, self.Q + 1)
            for qb in range(0, self.Q - qa + 1)
        ]

    def _label_wrongs_(self, q):
        return [
            (w1, w2) for w1 in range(0, q + 1) for w2 in range(0, q - w1 + 1)
        ]



"""Tests for the module ntqr.r2.alarms."""

import ntqr.r2.raxioms
import ntqr.alarms
import ntqr.r2.evaluations

import pytest

labels = ("a", "b")
c1Axioms = ntqr.r2.raxioms.SingleClassifierAxioms(labels, "1")
c2Axioms = ntqr.r2.raxioms.SingleClassifierAxioms(labels, "2")


# Better than 50%
factors = (2, 2)
qs = (4, 6)


@pytest.mark.parametrize(
    "qs, factors, corrects, satisfies",
    (
        (qs, factors, (3, 4), True),
        (qs, factors, (2, 4), False),
        (qs, factors, (3, 3), False),
    ),
)
def test_single_classifier_alarm_generate_safety_specification(
    qs, factors, corrects, satisfies
):
    safety_specification = ntqr.alarms.LabelsSafetySpecification(factors)
    assert safety_specification.is_satisfied(qs, corrects) == satisfies


alarm = ntqr.alarms.SingleClassifierAxiomAlarm(
    10,
    [c1Axioms, c2Axioms],
    ntqr.r2.evaluations.SingleClassifierEvaluations,
)
alarm.set_safety_specification(factors)


@pytest.mark.parametrize(
    "alarm, qs, factors, responses, misaligned",
    (
        (alarm, qs, factors, ((2, 8), (3, 7)), True),
        (alarm, qs, factors, ((4, 6), (3, 7)), False),
        (alarm, qs, factors, ((2, 8), (8, 2)), True),
    ),
)
def test_misalignment_alarm_at_qs(alarm, qs, factors, responses, misaligned):
    misalignment_test = alarm.misaligned_at_qs(qs, responses)
    assert misalignment_test == misaligned


alarm = ntqr.alarms.SingleClassifierAxiomAlarm(
    20,
    [c1Axioms, c2Axioms],
    ntqr.r2.evaluations.SingleClassifierEvaluations,
)
alarm.set_safety_specification(factors)


@pytest.mark.parametrize(
    "alarm, factors, responses, misaligned",
    (
        (alarm, factors, ((16, 4), (17, 3)), False),
        (alarm, factors, ((16, 4), (1, 19)), True),
    ),
)
def test_misalignment_alarm(alarm, factors, responses, misaligned):
    assert alarm.are_misaligned(responses) == misaligned



"""Tests for the ntqr.r2.evaluations module."""

import ntqr.r2.raxioms
import ntqr.r2.evaluations

labels = ["a", "b"]
classifier = "k"


r2axioms = ntqr.r2.raxioms.SingleClassifierAxioms(labels, classifier)
scEvaluations = ntqr.r2.evaluations.SingleClassifierEvaluations(10, r2axioms)


def test_errors_at_qs():
    evals = scEvaluations.errors_at_qs([3, 7], [5, 5])
    assert evals == set(
        [((0,), (2,)), ((1,), (3,)), ((2,), (4,)), ((3,), (5,))]
    )


def test_max_correct_at_qa_qb():
    max_correct = scEvaluations.max_correct_at_qs([3, 7], [5, 5])
    assert max_correct == [3, 5]



"""Module for convenience plotting functions related to NTQR algorithms."""

import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
import itertools

import ntqr.r2.raxioms
import ntqr.r3.raxioms
import ntqr.r2.evaluations
import ntqr.r3.evaluations
import ntqr.alarms


def plot_pair_logical_alarm_at_qs(labels, classifiers, qs, responses, plt=plt):

    num_labels = len(labels)
    assert len(classifiers) == 2
    assert len(classifiers) == len(responses)
    assert sum(responses[0]) == sum(responses[1])
    Q = sum(responses[0])

    # Select the right axiom and evaluation classes
    if num_labels == 2:
        axioms_cls = ntqr.r2.raxioms.SingleClassifierAxioms
        evals_cls = ntqr.r2.evaluations.SingleClassifierEvaluations
    elif num_labels == 3:
        axioms_cls = ntqr.r3.raxioms.SingleClassifierAxioms
        evals_cls = ntqr.r3.evaluations.SingleClassifierEvaluations

    classifier_axioms = [
        axioms_cls(labels, classifier) for classifier in classifiers
    ]
    classifier_vars = zip(
        *[
            [
                axioms.responses_by_label[label]["correct"][label]
                for label in labels
            ]
            for axioms in classifier_axioms
        ]
    )

    classifier_evals = [evals_cls(Q, axioms) for axioms in classifier_axioms]
    # print(classifier_evals)
    # print(responses[0])
    # print(classifier_evals[0].errors_at_qs(qs, responses[0]))

    classifier_evals_by_label = zip(
        *[
            zip(
                *list(
                    classifier_by_label_evals
                    for classifier_by_label_evals in eval.correct_at_qs(
                        qs, classifier_responses
                    )
                ),
                strict=True,
            )
            for eval, classifier_responses in zip(classifier_evals, responses)
        ]
    )

    safety_specification = ntqr.alarms.LabelsSafetySpecification(
        [2] * num_labels
    )
    label_safety_evals = safety_specification.pair_safe_evaluations_at_qs(qs)

    # Start building plot
    fig, axs = plt.subplots(1, num_labels)
    for ql, ax, pair_label_values, safety_evals, label_vars in zip(
        qs, axs, classifier_evals_by_label, label_safety_evals, classifier_vars
    ):
        if ql > 0:
            _plot_pair_label_evals(
                ax, ql, pair_label_values, safety_evals, label_vars
            )
        else:
            ax.axis("off")

    fig.tight_layout(pad=2.0)

    plt.show()


def plot_pair_label_evals(ql, pair_label_evals, safety_evals, vars, plt=plt):
    x, y = zip(*pair_label_evals)

    plt.rcParams["text.usetex"] = True

    fig, ax = plt.subplots()
    ax.set_aspect("equal")
    ax.set_xlabel("${}$".format(vars[0].name))
    ax.set_ylabel("${}$".format(vars[1].name), rotation="horizontal")
    ax.scatter(x, y, color="blue", label="evals")

    default_marker_size = plt.rcParams["lines.markersize"] ** 2

    xs, ys = zip(*safety_evals)
    ax.scatter(
        xs,
        ys,
        color="green",
        label="safety",
        marker="+",
        alpha=1.0,
        s=(default_marker_size + 90),
    )

    ax.xaxis.set_major_locator(MaxNLocator(integer=True))
    ax.yaxis.set_major_locator(MaxNLocator(integer=True))

    ax.set_xlim(0, ql)
    ax.set_ylim(0, ql)

    plt.show()


def _plot_pair_label_evals(
    ax, ql, pair_label_evals, safety_evals, vars, plt=plt
):

    # Here is where we do the product
    x, y = zip(*itertools.product(*pair_label_evals))

    plt.rcParams["text.usetex"] = True

    ax.set_aspect("equal")
    ax.set_xlabel("${}$".format(vars[0]))
    ax.set_ylabel("${}$".format(vars[1]), rotation="horizontal")
    ax.scatter(x, y, color="blue", label="evals")

    default_marker_size = plt.rcParams["lines.markersize"] ** 2

    xs, ys = zip(*safety_evals)
    ax.scatter(
        xs,
        ys,
        color="green",
        label="safety",
        marker="+",
        alpha=1.0,
        s=(default_marker_size + 3 * default_marker_size / 2),
    )

    ax.xaxis.set_major_locator(MaxNLocator(integer=True))
    ax.yaxis.set_major_locator(MaxNLocator(integer=True))

    extra_bit = 1 / 4
    ax.set_xlim(-extra_bit, ql + extra_bit)
    ax.set_ylim(-extra_bit, ql + extra_bit)

    return



"""@author: Andrés Corrada-Emmanuel."""
import math, itertools

# The mathematics of evaluating finite samples is, by construction, one
# of estimating integer fractions. We import this module so we can create
# two versions of every computation - the exact one using integer ratios,
# and the one using the default floating point numbers
import sympy
from typing_extensions import Union, Literal, Mapping, Iterable
from dataclasses import dataclass

# Types
Label = Union[Literal["a"], Literal["b"]]


# A vote is an ordered tuple of the decisions
Votes = tuple[Label, ...]

# Vote counts are what we see when we have unlabeled data.
VoteCounts = Mapping[Votes, int]

# Label vote counts, vote counts by true label, are only available
# when we are carrying out experiments on labeled data.
LabelVoteCounts = Mapping[Label, VoteCounts]

VoteFrequencies = Mapping[Votes, sympy.Rational]


# Some definitions and utilities to help various classes
opposite_label = {"a": "b", "b": "a"}


# To compute algebraic functions of classifier statistics,
# we need helper functions that pick out the vote patterns
# where a classifer voted with a given label.
def classifier_label_votes(
    classifier: int, label: Label, vote_patterns: Iterable[Votes]
) -> tuple[Votes, ...]:
    """
    All the vote patterns where classifier voted with label.

    Parameters
    ----------
    classifier : int
        Index of the classifier in the vote_patterns.
    label : Label
        The classifier label vote.

    Returns
    -------
    tuple[Votes, ...]
        All the vote patterns where classifier voted with label.
    """
    return tuple(
        [votes for votes in vote_patterns if votes[classifier] == label]
    )


def votes_match(
    votes: tuple[Label, ...],
    classifiers: Iterable[int],
    labels: Iterable[Label],
) -> bool:
    """Test if labels by classifiers matches votes.

    Parameters
    ----------
    votes : Tuple[Label, ...]
        Vote pattern for the trio.
    classifiers : Iterable[int]
        Classifiers to check.
    labels : Iterable[Label]
        Label voted by each classifier - the voting
        sub-pattern we want a trio pattern to match.

    Returns
    -------
    bool.

    """
    return all(
        [
            votes[classifier] == label
            for classifier, label in zip(classifiers, labels)
        ]
    )


def classifiers_labels_votes(
    classifiers: Iterable[int],
    labels: Iterable[Label],
    vote_patterns: Iterable[Votes],
) -> tuple[Votes, ...]:
    """Trio voting patterns that match labels by classifiers.

    Parameters
    ----------
    classifiers : Iterable[int]
        The indices of the classifiers in the trio
        to match.
    labels : Iterable[Label]
        Voting sub-pattern by classifiers to match.

    Returns
    -------
    tuple[Votes, ...]
        Tuple of trio voting patterns that matched labels
        for the classifiers.

    """
    return tuple(
        [
            votes
            for votes in vote_patterns
            if votes_match(votes, classifiers, labels)
        ]
    )


class TrioLabelVoteCounts:
    """
    Data class for the by-label aligned votes of three binary classifiers.

    This class is only useful in an experimental setting where one has
    observed a test with **labeled** data.
    Initialized with a Mapping[Label, Mapping[Votes, int]] of the form
    {
     'a':{('a', 'a', 'a'): int, ..., ('b', 'b', 'b'):int},
     'b':{('a', 'a', 'a'): int, ..., ('b', 'b', 'b'):int}
    }

    DEPRECATED: This clas is being replaced with the upcoming LabelVoteCounts
    for an arbitrary number of classifiers.
    """

    vote_patterns = list(itertools.product(*["ab" for i in range(3)]))
    pairs = ((0, 1), (0, 2), (1, 2))

    def __init__(self, label_vote_counts):
        """All labels and vote patterns are set."""

        self.label_vote_counts = {
            label: {
                votes: label_vote_counts.get(label, {}).get(votes, 0)
                for votes in self.vote_patterns
            }
            for label in ("a", "b")
        }

        # self.label_vote_counts = {"b":  1}

        object.__setattr__(
            self,
            "test_sizes",
            {
                label: sum(self.label_vote_counts[label].values())
                for label in ("a", "b")
            },
        )

    def to_vote_counts(self) -> VoteCounts:
        """
        Turn by-label counts into by-vote-pattern counts.

        Using {'a':{..., ('a', 'b', 'a'): x, ...},
               'b':{..., ('a', 'b', 'a'): y, ...}}

        Returns
        -------
        {..., ('a', 'b', 'a'): x+y, ...}
        """
        return {
            votes: sum(
                [self.label_vote_counts[label][votes] for label in ("a", "b")]
            )
            for votes in self.vote_patterns
        }

    def to_TrioVoteCounts(self):
        """Return TrioVoteCounts object by summing votes across labels."""
        return TrioVoteCounts(self.to_vote_counts())

    def to_voting_frequency_fractions(self) -> VoteFrequencies:
        """
        Compute observed voting pattern frequencies.

        Returns
        -------
        Mapping[Votes, Fraction]
        """
        by_voting_counts = self.to_vote_counts()
        size_of_test = sum(by_voting_counts.values())
        return {
            vp: sympy.Rational(by_voting_counts[vp], size_of_test)
            for vp in by_voting_counts.keys()
        }

    def to_voting_frequencies_float(self) -> Mapping[Votes, float]:
        """
        Compute observed voting frequencies inexactly, as floats.

        Returns
        -------
        Mapping[Votes, float]
        """
        by_voting_counts = self.to_vote_counts()
        size_of_test = sum(by_voting_counts.values())
        return {
            vp: by_voting_counts[vp] / size_of_test
            for vp in by_voting_counts.keys()
        }

    def __getitem__(self, label):
        """
        Return the vote pattern counts for label.

        Parameters
        ----------
        label :Label
            One of 'a' or 'b'.

        Returns
        -------
        Mapping[Votes, int]
            The aligned vote counts observed for the given label.
        """
        return self.label_vote_counts[label]

    def flip_classifiers_label_decisions(
        self, classifiers: Iterable, label: Label
    ):
        flipped_data_sketch = {}
        opp_label = opposite_label[label]
        flipped_data_sketch[opp_label] = self.label_vote_counts[opp_label]
        flipped_vote_counts = {}
        for vp, count in self.label_vote_counts[label].items():
            new_pattern = tuple(
                [
                    opposite_label[vp[i]] if i in classifiers else vp[i]
                    for i in range(3)
                ]
            )
            flipped_vote_counts[new_pattern] = count
        flipped_data_sketch[label] = flipped_vote_counts

        return TrioLabelVoteCounts(flipped_data_sketch)


@dataclass
class TrioVoteCounts:
    """Data class to validate the test counts for three binary classifiers.

    Initialized with a Mapping[Votes, int] of the form:
        {('a', 'a', 'a'): int, ..., ('b', 'b', 'b'):int}

    This is the class that is used for evaluation on unlabeled data where
    we only have access to the aligned decisions of the binary classifiers
    and have no knowledge of label of any one item that was classified.

    DEPRECATED: This class will be replaced with the ObservedVoteCounts class
    that can handle an arbitrary number of classifiers.
    """

    # Three binary classifiers have eight possible voting patterns
    vote_patterns = list(itertools.product(*["ab" for i in range(3)]))

    pairs = ((0, 1), (0, 2), (1, 2))

    vote_counts: VoteCounts

    def __post_init__(self):
        """
        Check we have counts for a valid evaluation of binary classifiers.

        1. No negative counts.
        2. Initialize all possible vote patterns by the trio.
        3. The empty test - all counts zero - is not allowed.
        """
        # Make sure all patterns have a non-negative count.
        object.__setattr__(
            self,
            "vote_counts",
            {
                vote_pattern: self.vote_counts.get(vote_pattern, 0)
                for vote_pattern in self.vote_patterns
            },
        )

        if any([count for count in self.vote_counts.values() if (count < 0)]):
            raise ValueError("No negative vote counts allowed.")

        object.__setattr__(self, "test_size", sum(self.vote_counts.values()))

        # The empty test is not allowed
        if self.test_size == 0:
            raise ValueError("The empty test is not allowed.")

    def to_frequencies_exact(self) -> VoteFrequencies:
        """
        Turn vote integer counts to exact Fraction objects.

        Returns
        -------
        VoteFrequencies:
            Maps a trio vote pattern to its Fraction occurence in the test.
        """
        return {
            vp: sympy.Rational(self.vote_counts[vp], self.test_size)
            for vp in self.vote_counts.keys()
        }

    def to_frequencies_float(self) -> Mapping[Votes, float]:
        """
        Compute observerd voting pattern frequencies, inexactly, as floats.

        Returns
        -------
        Mapping[Votes, float]:
            Maps a trio vote pattern to its percentage occurence in the test
            as an inexact float.

        """
        return {
            vp: self.vote_counts[vp] / self.test_size
            for vp in self.vote_counts.keys()
        }

    def classifier_label_frequency(
        self, classifier: int, label: Label
    ) -> sympy.Rational:
        """
        Calculate classifier label voting frequency.

        Parameters
        ----------
        classifier : int
            The index of the classifier.
        label : Label
            The label.

        Returns
        -------
        sympy.Rational(label_vote_counts, test_size):
            The fraction of times the classifier voted the label when
            classifying items in the test.

        """
        vote_frequencies = self.to_frequencies_exact()
        return sum(
            [
                vote_frequencies[votes]
                for votes in classifier_label_votes(
                    classifier, label, self.vote_patterns
                )
            ]
        )

    def classifier_label_responses(self, classifier: int, label: Label) -> int:
        """
        Calculates number of responses with label by classifier.

        Parameters
        ----------
        classifier : int
            DESCRIPTION.
        label : Label
            DESCRIPTION.

        Returns
        -------
        int
            Number of times the classifier decided an item was label.

        """
        return sum(
            [
                self.vote_counts[votes]
                for votes in classifier_label_votes(
                    classifier, label, self.vote_patterns
                )
            ]
        )

    def pair_label_frequency(
        self, pair: Iterable[int], label: Label
    ) -> sympy.Rational:
        """
        Compute frequency of times a pair voted with the same label.

        Parameters
        ----------
        pair : Iterable[int, int]
            Classifier indicies.
        label : Label
            The label.

        Returns
        -------
        sympy.Rational:
            The fraction of times a pair of classifiers voted with the
            same label when classifying items in the test.

        """
        vote_frequencies = self.to_frequencies_exact()
        return sum(
            [
                vote_frequencies[votes]
                for votes in classifiers_labels_votes(
                    pair, (label, label), self.vote_patterns
                )
            ]
        )

    def pair_label_responses(
        self, pair: Iterable[int], label: Label
    ) -> sympy.Rational:
        """
        Computes number of times a pair voted with the same label.

        Parameters
        ----------
        pair : Iterable[int, int]
            Classifier indicies.
        label : Label
            The label.

        Returns
        -------
        int:
            Number of items pair voted with the same label.

        """
        return sum(
            [
                self.vote_counts[votes]
                for votes in classifiers_labels_votes(
                    pair, (label, label), self.vote_patterns
                )
            ]
        )

    def pair_frequency_moment(
        self, pair: Iterable[int], label: Label
    ) -> sympy.Rational:
        """
        Calculate the label classifier pair frequency moment.

        If (i, j) = pair, then this is -

            f_{label_i, label_j} - f_{label_i} * f_{label_j}

        The fraction of times the classifier pair voted with the same label
        minus the product of their individual label voting frequencies.

        Parameters
        ----------
        pair : Iterable(int, int)
            The pair of classifiers.
        label : Label
            One of the binary labels.

        Returns
        -------
        sympy.Rational:
            The pair frequency moment as a Fraction object
        """
        label_frequencies = [
            self.classifier_label_frequency(classifier, label)
            for classifier in pair
        ]
        return self.pair_label_frequency(pair, label) - math.prod(
            label_frequencies
        )

    def label_pairs_frequency_moments(
        self, label: Label
    ) -> Mapping[tuple[int, int], sympy.Rational]:
        """All the label pair frequency moments."""
        return {
            pair: self.pair_frequency_moment(pair, label)
            for pair in self.pairs
        }

    def trio_frequency_moment(self) -> sympy.Rational:
        """
        Calculate the 3rd frequency moment of a trio of binary classifiers.

        Don't ask.
        """
        classifier_label_frequencies = [
            self.classifier_label_frequency(classifier, "b")
            for classifier in range(3)
        ]
        prod_frequencies = math.prod(classifier_label_frequencies)
        pfmds = self.label_pairs_frequency_moments("b")
        sum_prod = (
            classifier_label_frequencies[0] * pfmds[(1, 2)]
            + classifier_label_frequencies[1] * pfmds[(0, 2)]
            + classifier_label_frequencies[2] * pfmds[(0, 1)]
        )
        return prod_frequencies + sum_prod


@dataclass
class ObservedVoteCounts:
    """Data class to validate the test vote counts for an arbitrary number
    of binary classifiers.

    Initialized with a Mapping[Votes, int] of the form:
        {('a', ..., a'): int, ..., ('b', ..., 'b'):int}

    Class used during evaluations with unlabeled data where we do not
    know the true label for each item labeled by the classifiers.
    """

    vote_counts: VoteCounts

    def __post_init__(self):
        """
        Check we have counts for a valid evaluation of binary classifiers.

        0. Determine the size of the ensemble and that all provided
           vote patterns conform to it.
        1. No negative counts.
        2. Initialize all possible vote patterns by the trio.
        3. The empty test - all counts zero - is not allowed.
        """

        # Start by determining the size of the ensemble is the same
        # for all provided vote patterns
        ensemble_sizes = [
            len(vote_pattern) for vote_pattern in self.vote_counts.keys()
        ]
        assert len(set(ensemble_sizes)) == 1

        # Construct the binary vote patterns for an ensemble
        # of this size
        self.ensemble_size = ensemble_sizes[0]
        self.vote_patterns = itertools.product(
            *["ab" for i in range(self.ensemble_size)]
        )

        # Make sure all patterns have a non-negative count.
        object.__setattr__(
            self,
            "vote_counts",
            {
                vote_pattern: self.vote_counts.get(vote_pattern, 0)
                for vote_pattern in self.vote_patterns
            },
        )

        if any([count for count in self.vote_counts.values() if (count < 0)]):
            raise ValueError("No negative vote counts allowed.")

        object.__setattr__(self, "test_size", sum(self.vote_counts.values()))

        # The empty test is not allowed
        if self.test_size == 0:
            raise ValueError("The empty test is not allowed.")

    def to_frequencies_exact(self) -> VoteFrequencies:
        """
        Turn vote integer counts to exact Fraction objects.

        Returns
        -------
        VoteFrequencies:
            Maps a trio vote pattern to its Fraction occurence in the test.
        """
        return {
            vp: sympy.Rational(self.vote_counts[vp], self.test_size)
            for vp in self.vote_counts.keys()
        }

    def to_frequencies_float(self) -> Mapping[Votes, float]:
        """
        Compute observerd voting pattern frequencies, inexactly, as floats.

        Returns
        -------
        Mapping[Votes, float]:
            Maps a trio vote pattern to its percentage occurence in the test
            as an inexact float.

        """
        return {
            vp: self.vote_counts[vp] / self.test_size
            for vp in self.vote_counts.keys()
        }

    def classifier_label_frequency(
        self, classifier: int, label: Label
    ) -> sympy.Rational:
        """
        Calculate classifier label voting frequency.

        Parameters
        ----------
        classifier : int
            The index of the classifier.
        label : Label
            The label.

        Returns
        -------
        sympy.Rational(label_vote_counts, test_size):
            The fraction of times the classifier voted the label when
            classifying items in the test.

        """
        vote_frequencies = self.to_frequencies_exact()
        return sum(
            [
                vote_frequencies[votes]
                for votes in classifier_label_votes(classifier, label)
            ]
        )

    def classifier_label_responses(self, classifier: int, label: Label) -> int:
        """
        Calculates number of responses with label by classifier.

        Parameters
        ----------
        classifier : int
            DESCRIPTION.
        label : Label
            DESCRIPTION.

        Returns
        -------
        int
            Number of times the classifier decided an item was label.

        """
        return sum(
            [
                self.vote_counts[votes]
                for votes in classifier_label_votes(classifier, label)
            ]
        )

    def pair_label_frequency(
        self, pair: Iterable[int], label: Label
    ) -> sympy.Rational:
        """
        Compute frequency of times a pair voted with the same label.

        Parameters
        ----------
        pair : Iterable[int, int]
            Classifier indicies.
        label : Label
            The label.

        Returns
        -------
        sympy.Rational:
            The fraction of times a pair of classifiers voted with the
            same label when classifying items in the test.

        """
        vote_frequencies = self.to_frequencies_exact()
        return sum(
            [
                vote_frequencies[votes]
                for votes in classifiers_labels_votes(pair, (label, label))
            ]
        )

    def pair_label_responses(
        self, pair: Iterable[int], label: Label
    ) -> sympy.Rational:
        """
        Computes number of times a pair voted with the same label.

        Parameters
        ----------
        pair : Iterable[int, int]
            Classifier indicies.
        label : Label
            The label.

        Returns
        -------
        int:
            Number of items pair voted with the same label.

        """
        return sum(
            [
                self.vote_counts[votes]
                for votes in classifiers_labels_votes(pair, (label, label))
            ]
        )

    def pair_frequency_moment(
        self, pair: Iterable[int], label: Label
    ) -> sympy.Rational:
        """
        Calculate the label classifier pair frequency moment.

        If (i, j) = pair, then this is -

            f_{label_i, label_j} - f_{label_i} * f_{label_j}

        The fraction of times the classifier pair voted with the same label
        minus the product of their individual label voting frequencies.

        Parameters
        ----------
        pair : Iterable(int, int)
            The pair of classifiers.
        label : Label
            One of the binary labels.

        Returns
        -------
        sympy.Rational:
            The pair frequency moment as a Fraction object
        """
        label_frequencies = [
            self.classifier_label_frequency(classifier, label)
            for classifier in pair
        ]
        return self.pair_label_frequency(pair, label) - math.prod(
            label_frequencies
        )

    def label_pairs_frequency_moments(
        self, label: Label
    ) -> Mapping[tuple[int, int], sympy.Rational]:
        """All the label pair frequency moments."""
        return {
            pair: self.pair_frequency_moment(pair, label)
            for pair in trio_pairs
        }

    def trio_frequency_moment(self) -> sympy.Rational:
        """
        Calculate the 3rd frequency moment of a trio of binary classifiers.

        Don't ask.
        """
        classifier_label_frequencies = [
            self.classifier_label_frequency(classifier, "b")
            for classifier in range(3)
        ]
        prod_frequencies = math.prod(classifier_label_frequencies)
        pfmds = self.label_pairs_frequency_moments("b")
        sum_prod = (
            classifier_label_frequencies[0] * pfmds[(1, 2)]
            + classifier_label_frequencies[1] * pfmds[(0, 2)]
            + classifier_label_frequencies[2] * pfmds[(0, 1)]
        )
        return prod_frequencies + sum_prod



"""R=2 evaluation axioms in integer space.

This is the space of integers possible since we are dealing with finite
tests. Statistics of correctness are given as integers. In the case of
R=2 the evaluation of a single classifier/responder is (Q_a, R_a_a, R_b_b).
"""

import sympy
from ntqr.statistics import SingleClassifierVariables


class SingleClassifierAxioms:
    """
    The axioms of a single R=2 classifier

    ...

    Attributes
    ----------

    question_numbers : List[simpy.Symbol..]
        The variables for the count of correct questions for each true
        label.

    responses : List[simpy.Symbol..]
        The variables for the observed interger count of the two classes.
        Generically of the form,

        R_{l_i} : number of 'l' responses by classifier 'i'

    correctness_variables: List[simpy.Symbol..]
        The variables associated with correct and wrong responses given
        the true label. Generically of the form,

        R_{l_i, l_true}: number of 'l' responses by classifier 'i' given
        true label 'l_true'.


    Methods
    -------
    evaluate_axioms(eval_dict): List[expression, expression, expression]
        Evaluates the axioms given the variable substitutions in 'eval_dict'.
    satisfies_axioms(eval_dict): Boolean
        Checks if the variable substitutions in 'eval_dict' make all three
        axioms identically zero.
    """

    def __init__(self, labels, classifier):
        "Constructs variables for 'labels' and the axioms they must satisfy."

        self.labels = labels
        self.classifier = classifier

        vars = SingleClassifierVariables(labels, classifier)
        self.questions_number = vars.questions_number
        self.responses = vars.responses
        self.responses_by_label = vars.responses_by_label

        # Construct the three, dependent axioms for a single classifier
        self.algebraic_expressions = {}
        for i_true in range(len(labels)):

            true_label = labels[i_true]
            true_label_responses = self.responses_by_label[true_label]
            q_number = self.questions_number[true_label]

            mistakes_out_of_label = q_number
            mistakes_into_label = 0
            for i_response in range(len(labels)):
                response_label = labels[i_response]
                if i_response != i_true:
                    mistakes_out_of_label -= true_label_responses[
                        response_label
                    ]
                    mistakes_into_label += self.responses_by_label[
                        response_label
                    ][true_label]

            self.algebraic_expressions[true_label] = (
                mistakes_out_of_label
                + mistakes_into_label
                - self.responses[true_label]
            )

    def evaluate_axioms(self, eval_dict):
        """
        Evaluates axioms given 'eval_dict'.

        Parameters
        ----------
        eval_dict: Map[sympySymbol -> value]

        Returns
        -------
        Dict mapping label to axiom expression.
        """

        return {
            label: axiom.subs(eval_dict)
            for label, axiom in self.algebraic_expressions.items()
        }

    def satisfies_axioms(self, eval_dict):
        """
        Tests axioms are satisfied for 'eval_dict' values.

        Parameters
        ----------
        eval_dict: Map[sympySymbol -> value]

        Returns
        -------
        Boolean: returns True if the axioms are identically zero, False
        otherwise.
        """

        evaluated_axioms = self.evaluate_axioms(eval_dict)
        return all([(axiom == 0) for axiom in evaluated_axioms.values()])


## The following code needs to be refactored into classes for
## pair and triplet axioms
# Symbols and axioms for pairs
raja = sympy.Symbol(r"R_{a_j,a}")
rbjb = sympy.Symbol(r"R_{b_j,b}")

raka = sympy.Symbol(r"R_{a_k,a}")
rbkb = sympy.Symbol(r"R_{b_k,b}")

raiaja = sympy.Symbol(r"R_{a_i, a_j; a}")
raiaka = sympy.Symbol(r"R_{a_i, a_k; a}")
rajaka = sympy.Symbol(r"R_{a_j, a_k; a}")

rbibjb = sympy.Symbol(r"R_{b_i, b_j; b}")
rbibkb = sympy.Symbol(r"R_{b_i, b_k; b}")
rbjbkb = sympy.Symbol(r"R_{b_j, b_k; b}")

# pair_binary_responders_axiom = [
#     (qb - (rai + raj) + raiaj) + ((raia + raja) - (raiaja + rbibjb)),
#     (qa - (rbi + rbj) + rbibj) + ((rbib + rbjb) - (raiaja + rbibjb)),
# ]
three_responders_axiom = []



"""Algorithms for logical alarms based on the axioms.

Formal verification of unsupervised evaluations is carried out by
using the agreements and disagreements between classifiers to detect
if they are misaligned given a safety specification.

The 'atomic' logical test for the alarms is a look at the group
evaluations that are logically consistent with how the classifiers
aligned in their decisions and an assumed number of corrects for
each label in the true, but unknown, answer key for the exam.

For example, in a test with three possible responses or classes
for each question, we need to specify,

    qs = (q_label_1, q_label_2, q_label_3)

where

    sum(qs) = Q,

with Q the size of the test. So a test with Q=10, could have a qs setting of, 
(5,3,2) since sum(5,3,2) = 10.

This atomic misalignment test at fixed qs value then allows you
to create custom alarms depending on your application domain.
Some examples,

1. The prevalence of classes in your tests is biased toward
small amounts of one label or the other. In that case, you can construct
an alarm as,

        and([alarm.misaligned_at_qs(qs, responses) for qs my_range()])

2. The method `ntqr.SingleClassifierAxiomsAlarm.are_misaligned`
is a test for fully unsupervised settings and is equivalent to,

        and([alarm.misaligned_at_qs((qa,Q-qa), rs) for qa in range(0,Q+1)])

That is, the only thing you have are the classifier responses and the
size of the test, Q.

3. You believe that your classifiers are high performing and therefore
will only accept (Q_label_1, Q_label_2, ...) settings for which
all your classifiers are better than x% at detecting all the labels.
This turns the atomic logical test into a measuring instrument for
the prevalence of the labels in the tested dataset.
"""

import itertools


class SingleClassifierAxiomsAlarm:
    """Alarm based on the single classifier axioms for the ensemble members

    Initialize with 'classifiers_axioms', list of
    ntqr.r3.raxioms.SingleClassifierAxioms, one per classifier we want
    to compare.

    Also needs class that does SingleClassifierEvaluations given the
    axioms as 'cls_single_evals'
    """

    def __init__(self, Q, classifiers_axioms, cls_single_evals):
        """"""

        self.Q = Q

        self.classifiers_axioms = classifiers_axioms
        self.labels = classifiers_axioms[0].labels

        self.evals = [
            cls_single_evals(self.Q, axioms)
            for axioms in self.classifiers_axioms
        ]

    def set_safety_specification(self, factors):
        """Set alarm's safetySpecification given factors"""
        assert len(factors) == len(self.labels)
        self.safety_specification = LabelsSafetySpecification(factors)

    def misaligned_at_qs(self, qs, responses):
        """Boolean test to see if the classifiers violate the safety
        specification at given questions correct number.
        """

        assert self.check_responses(qs, responses)

        max_corrects = [
            sca_eval.max_correct_at_qs(qs, cresponses)
            for sca_eval, cresponses in zip(self.evals, responses)
        ]

        return any(
            (not self.safety_specification.is_satisfied(qs, max_correct))
            for max_correct in max_corrects
        )

    def misalignment_trace(self, responses):
        "Boolean trace of the classifiers test alignment."
        all_qs = self.evals[0].all_qs()
        trace = set(
            (qs, self.misaligned_at_qs(qs, responses)) for qs in all_qs
        )
        return trace

    def are_misaligned(self, responses):
        "Boolean test of misaligned"
        trace = self.misalignment_trace(responses)
        return all(list(zip(*trace))[1])

    def check_responses(self, qs, responses):
        "Checks logical constraints on responses."
        q = sum(qs)
        if q == self.Q:
            qs_check = True
        else:
            qs_check = False
        responses_check = all(
            [
                (q == sum(classifier_responses))
                for classifier_responses in responses
            ]
        )

        consistency_check = all([qs_check, responses_check])
        if consistency_check == False:
            print("q ", q, "Q ", self.Q),
            print("qs ", qs)
            print("responses ", responses)
            print("checks ", [qs_check, responses_check])

        return all([qs_check, responses_check])


class LabelsSafetySpecification:
    """Simple example of a safety specification.
    Parameters:
    ----------
    factors: list of factors to be used in safety
    specification tests, one per label.
    """

    def __init__(self, factors):

        self.factors = factors

    def is_satisfied(self, qs, correct_responses):
        """Checks that list of label accuracies satisfy the
        safety specification.

        Returns True if all labels satisfy the equation
          factor*correct_response - ql > 0
        given each label's factor and assumed number in
        the unknown test answer key, False otherwise.
        """
        tests = [
            (factor * correct_response - ql) > 0 if ql > 0 else True
            for factor, correct_response, ql in zip(
                self.factors, correct_responses, qs
            )
        ]
        return all(tests)

    def pair_safe_evaluations_at_qs(self, qs):
        """All pair evaluations satisfying safety spec at given qs."""
        correct_ranges = [
            [
                q_correct
                for q_correct in range(0, ql + 1)
                if ((factor * q_correct) - ql > 0)
            ]
            for factor, ql in zip(self.factors, qs)
        ]

        return [
            itertools.product(correct_range, repeat=2)
            for correct_range in correct_ranges
        ]


class GradeSafetySpecification:
    """Simple example of a safety specification.
    Parameters:
    -----------
    factors: list of factors to be used in safety
    specification tests, one per label.

    Returns:
    --------
    True if all labels satisfy factor_l*max_correct_l - ql > 0,
    False otherwise
    """

    def __init__(self, factors):

        self.factors = factors

    def is_satisfied(self, qs, correct_responses):
        """Checks that list of label accuracies satisfy the
        safety specification.

        Returns True if the max correct answers for labels satisfy
          factor*sum(correct_label) - Q > 0 given questions
          numbers 'qs',
          False otherwise.
        """
        Q = sum(qs)
        grade_test = self.factor * sum(correct_responses) - Q > 0
        return grade_test

    def pair_safe_evaluations_at_qs(self, qs):
        """All pair evaluations satisfying safety spec at given qs."""
        ql_ranges = (
            [q_correct for q_correct in range(0, ql + 1)] for ql in qs
        )
        pair_safe_evaluations = filter(
            lambda rs: self.is_satisfied(qs, rs), itertools.product(*ql_ranges)
        )
        return [
            itertools.product(correct_range, repeat=2)
            for correct_range in zip(*pair_safe_evaluations)
        ]
